#ifndef __File_h__
#define __File_h__

#include "Error.h"
#include "HeapData.h"
#include "IFile.h"
#include "String.h"
#include "T/sys/stat.h"
#include "T/sys/types.h"
#include "T/unistd.h"

class File : public IFile
{
public:
    File( const String &path ) : _path( path ), _descriptor( NO_DESCRIPTOR )
    {
    }

    ~File()
    {
        closeIfNeeded();
    }

    static bool exists( const String &path )
    {
        struct stat DONT_CARE;
        return T::stat( path.ascii(), &DONT_CARE ) == 0;
    }

    static bool directory( const String &path )
    {
        struct stat fileData;
        memset( &fileData, 0, sizeof(struct stat) );

        if ( T::stat( path.ascii(), &fileData ) == 0 )
        {
            return S_ISDIR( fileData.st_mode );
        }

        return false;
    }

    static unsigned getSize( const String &path )
    {
        struct stat dataBuffer;
        memset( &dataBuffer, 0, sizeof(dataBuffer) );

        if ( T::stat( path.ascii(), &dataBuffer ) != 0 )
        {
            throw Error( Error::STAT_FAILED );
        }

        return dataBuffer.st_size;
    }

    void open( Mode mode )
    {
        if ( mode == IFile::MODE_READ )
        {
            openForReading();
        }
        else if ( mode == IFile::MODE_WRITE )
        {
            openForWriting();
        }
        else
        {
            openForAppending();
        }
    }

    void write( const String &line )
    {
        if ( T::write( _descriptor, line.ascii(), line.length() ) != (int)line.length() )
        {
            throw Error( Error::WRITE_FAILED );
        }
    }

    void write( const ConstSimpleData &data )
    {
        if ( T::write( _descriptor, data.data(), data.size() ) != (int)data.size() )
        {
            throw Error( Error::WRITE_FAILED );
        }
    }

    void read( HeapData &buffer, unsigned maxReadSize )
    {
        char readBuffer[maxReadSize];
        int bytesRead;

        if ( ( bytesRead = T::read( _descriptor, readBuffer, sizeof(readBuffer) ) ) == -1 )
        {
            throw Error( Error::READ_FAILED );
        }

        buffer = ConstSimpleData( readBuffer, bytesRead );
    }

    String readLine()
    {
        char byte;
        String result;

        while ( ( T::read( _descriptor, &byte, sizeof(byte) ) ) == 1 )
        {
            if ( byte == '\n' )
            {
                return result;
            }
            else
            {
                result += String( &byte, sizeof(byte) );
            }
        }

        return result;
    }

private:
    enum {
        NO_DESCRIPTOR = -1,
    };

    String _path;
    int _descriptor;

    void closeIfNeeded()
    {
        if ( _descriptor != NO_DESCRIPTOR )
        {
            T::close( _descriptor );
            _descriptor = NO_DESCRIPTOR;
        }
    }

    void openForWriting()
    {
        if ( ( _descriptor = T::open( _path.ascii(), O_RDWR | O_CREAT | O_TRUNC, S_IRUSR | S_IWUSR ) ) == NO_DESCRIPTOR )
        {
            throw Error( Error::OPEN_FAILED );
        }
    }

    void openForAppending()
    {
        if ( ( _descriptor = T::open( _path.ascii(), O_RDWR | O_CREAT | O_APPEND, S_IRUSR | S_IWUSR | S_IROTH | S_IRGRP ) ) == NO_DESCRIPTOR )
        {
            throw Error( Error::OPEN_FAILED );
        }
    }

    void openForReading()
    {
        mode_t DONT_CARE = (mode_t)NULL;

        if ( ( _descriptor = T::open( _path.ascii(), O_RDONLY, DONT_CARE ) ) == NO_DESCRIPTOR )
        {
            throw Error( Error::OPEN_FAILED );
        }
    }
};

#endif // __File_h__

//
// Local Variables:
// compile-command: "make -C .. "
// tags-file-name: "../TAGS"
// c-basic-offset: 4
// End:
//
