#ifndef __ShellCommandRunner_h__
#define __ShellCommandRunner_h__

#include "Error.h"
#include "String.h"
#include "T/sys/wait.h"
#include "T/unistd.h"

class ShellCommandRunner
{
public:
    ShellCommandRunner( const String &command ) : _command( command )
    {
    }

    void run( int &returnCode, String &output )
    {
        createPipe();
        fork();

        if ( returnedAsChild() )
        {
            closeReadEndOfPipe();
            redirectOutput();
            runShellCommand();
        }
        else
        {
            closeWriteEndOfPipe();
            readResult();
            waitForChild();
        }

        prepareOutput( returnCode, output );
    }

private:
    enum {
        PIPE_READ_DESCRIPTOR = 0,
        PIPE_WRITE_DESCRIPTOR = 1,
    };

    String _command;
    pid_t _returnedPid;
    int _pipeDescriptors[2];
    String _result;
    int _status;

    void createPipe()
    {
        if ( T::pipe( _pipeDescriptors ) == -1 )
        {
            throw Error( Error::PIPE_FAILED );
        }
    }

    void fork()
    {
        if ( ( _returnedPid = T::fork() ) < 0 )
        {
            throw Error( Error::FORK_FAILED );
        }
    }

    bool returnedAsChild() const
    {
        return _returnedPid == 0;
    }

    void closeReadEndOfPipe()
    {
        T::close( _pipeDescriptors[PIPE_READ_DESCRIPTOR] );
    }

    void closeWriteEndOfPipe()
    {
        T::close( _pipeDescriptors[PIPE_WRITE_DESCRIPTOR] );
    }

    void readResult()
    {
        enum {
            BUFFER_SIZE = 2048,
        };

        char buffer[BUFFER_SIZE];

        int bytesRead;

        while ( ( bytesRead = T::read( _pipeDescriptors[PIPE_READ_DESCRIPTOR], buffer, BUFFER_SIZE ) ) > 0 )
        {
            _result += String( buffer, bytesRead );
        }

        if ( bytesRead < 0 )
        {
            throw Error( Error::READ_FAILED );
        }
    }

    void redirectOutput()
    {
        if ( T::dup2( _pipeDescriptors[PIPE_WRITE_DESCRIPTOR], STDOUT_FILENO ) == -1 )
        {
            throw Error( Error::DUP_FAILED );
        }
    }

    void runShellCommand()
    {
        Vector<String> tokens = _command.tokenize( " " );

        char **argv = new char *[tokens.size() + 1];
        for ( unsigned i = 0; i < tokens.size(); ++i )
        {
            argv[i] = new char[tokens[i].length() + 1];
            strcpy( argv[i], tokens[i].ascii() );
        }

        argv[tokens.size()] = NULL;
        T::execvp( argv[0], argv );
    }

    void waitForChild()
    {
        static const int DONT_CARE_OPTIONS = 0;

        if ( T::waitpid( _returnedPid, &_status, DONT_CARE_OPTIONS ) == -1 )
        {
            throw Error( Error::WAITPID_FAILED );
        }
    }

    void prepareOutput( int &returnCode, String &output )
    {
        output = _result;
        returnCode = WEXITSTATUS( _status );
    }
};

#endif // __ShellCommandRunner_h__

//
// Local Variables:
// compile-command: "make -C .. "
// tags-file-name: "../TAGS"
// c-basic-offset: 4
// End:
//
