#include <cxxtest/TestSuite.h>

#include "Error.h"
#include "HeapData.h"
#include "MockErrno.h"
#include "MockForRawSocket.h"
#include "RawSocket.h"
#include "RealMalloc.h"

class RawSocketTestSuite : public CxxTest::TestSuite
{
public:
	MockForRawSocket *mock;

    void setUp()
	{
		TS_ASSERT( mock = new MockForRawSocket );
    }

	void tearDown()
	{
        TS_ASSERT_THROWS_NOTHING( delete mock );
	}

	void test_construcor_and_activate()
	{
		RawSocket *rawSocket;

		TS_ASSERT( rawSocket = new RawSocket( "127.0.0.1" ) );

        TS_ASSERT( !mock->socketWasCalled );

        TS_ASSERT_THROWS_NOTHING( rawSocket->activate() );

        TS_ASSERT( mock->socketWasCalled );

		TS_ASSERT( mock->bindWasCalled );

        sockaddr_in expectedBindSockaddr;
		memset( &expectedBindSockaddr, 0, sizeof(sockaddr_in ) );

		expectedBindSockaddr.sin_family = AF_INET;
        inet_aton( "127.0.0.1", &expectedBindSockaddr.sin_addr );

        TS_ASSERT_SAME_DATA( &mock->lastBindSockaddr, &expectedBindSockaddr, sizeof(sockaddr_in) );

        TS_ASSERT_THROWS_NOTHING( delete rawSocket );

		TS_ASSERT( mock->closeWasCalled );
    }

	void test_constructor_and_activate__socket_fails()
	{
        RawSocket *rawSocket;

		mock->nextSocketDescriptor = -1;

        TS_ASSERT( rawSocket = new RawSocket( "127.0.0.1" ) );

        TS_ASSERT_THROWS_EQUALS( rawSocket->activate(),
								 const Error &e,
								 e.code(),
								 Error::SOCKET_CREATION_FAILED );

        TS_ASSERT( !mock->closeWasCalled );
	}

	void test_constructor__bad_address()
	{
        RawSocket *rawSocket;

        mock->nextBindResult = -1;

        TS_ASSERT( rawSocket = new RawSocket( "127.0.0.1.bla" ) );

        TS_ASSERT_THROWS_EQUALS( rawSocket->activate(),
                                 const Error &e,
								 e.code(),
								 Error::INET_ATON_FAILED );

		TS_ASSERT( mock->closeWasCalled );
	}

	void test_constructor__bind_fails()
	{
        RawSocket *rawSocket;

        mock->nextBindResult = -1;

        TS_ASSERT( rawSocket = new RawSocket( "127.0.0.1" ) );

        TS_ASSERT_THROWS_EQUALS( rawSocket->activate(),
                                 const Error &e,
								 e.code(),
								 Error::BIND_FAILED );

		TS_ASSERT( mock->closeWasCalled );
	}

	void test_close()
	{
		RawSocket *rawSocket;

        TS_ASSERT( rawSocket = new RawSocket( "127.0.0.1" ) );
        TS_ASSERT_THROWS_NOTHING( rawSocket->activate() );

		TS_ASSERT( !mock->closeWasCalled );

        TS_ASSERT_THROWS_NOTHING( rawSocket->close() );

		TS_ASSERT( mock->closeWasCalled );

        mock->closeWasCalled = false;

        TS_ASSERT_THROWS_NOTHING( delete rawSocket );

		TS_ASSERT( !mock->closeWasCalled );
	}

    void test_send()
    {
		RawSocket *rawSocket;

        TS_ASSERT( rawSocket = new RawSocket( "127.0.0.1" ) );

        TS_ASSERT_THROWS_NOTHING( rawSocket->activate() );

        char buffer[] = { 0x0a, 0x0b, 0x0c, 0x0d, 0x0e };

        TS_ASSERT_THROWS_NOTHING( rawSocket->send( ConstSimpleData( buffer, sizeof(buffer) ), "182.32.11.22" ) );

        TS_ASSERT_EQUALS( mock->lastSentData.size(), sizeof(buffer) );
        TS_ASSERT_SAME_DATA( mock->lastSentData.data(), buffer, sizeof(buffer) );

        struct sockaddr_in expectedAddress;

        memset( &expectedAddress, 0, sizeof(expectedAddress) );
        inet_aton( "182.32.11.22", &expectedAddress.sin_addr );

        TS_ASSERT_SAME_DATA( &expectedAddress, &mock->lastSendAddress, sizeof(struct sockaddr_in) );

        TS_ASSERT_THROWS_NOTHING( delete rawSocket );
    }

    void test_send__fails()
    {
		RawSocket *rawSocket;

        TS_ASSERT( rawSocket = new RawSocket( "127.0.0.1" ) );

        TS_ASSERT_THROWS_NOTHING( rawSocket->activate() );

        mock->sendShouldFail = true;

        char buffer[] = { 0x0a, 0x0b, 0x0c, 0x0d, 0x0e };

        TS_ASSERT_THROWS_EQUALS( rawSocket->send( ConstSimpleData( buffer, sizeof(buffer) ), "182.32.11.22" ),
                                 const Error &e,
                                 e.code(),
                                 Error::SEND_FAILED );

        TS_ASSERT_THROWS_NOTHING( delete rawSocket );

    }

    void test_receive()
    {
		RawSocket *rawSocket;

        TS_ASSERT( rawSocket = new RawSocket( "127.0.0.1" ) );

        TS_ASSERT_THROWS_NOTHING( rawSocket->activate() );

        char buffer[] = { 0x0a, 0x0b, 0x0c, 0x0d, 0x0e };
        mock->nextReceiveData = ConstSimpleData( buffer, sizeof(buffer) );

        inet_aton( "82.32.11.2", &mock->nextSourceAddress.sin_addr );
        mock->nextSourceAddress.sin_port = 5588;

        HeapData data;
        struct sockaddr_in sourceAddress;

        TS_ASSERT_THROWS_NOTHING( rawSocket->receive( data, sourceAddress ) );

        TS_ASSERT_EQUALS( data.size(), sizeof(buffer) );
        TS_ASSERT_SAME_DATA( data.data(), buffer, sizeof(buffer) );

        TS_ASSERT_SAME_DATA( &mock->nextSourceAddress, &sourceAddress, sizeof(struct sockaddr_in) );

        TS_ASSERT_THROWS_NOTHING( delete rawSocket );
    }

    void test_receive__recv_fails()
    {
		RawSocket *rawSocket;

        TS_ASSERT( rawSocket = new RawSocket( "127.0.0.1" ) );

        TS_ASSERT_THROWS_NOTHING( rawSocket->activate() );

        char buffer[] = { 0x0a, 0x0b, 0x0c, 0x0d, 0x0e };
        mock->nextReceiveData = ConstSimpleData( buffer, sizeof(buffer) );

        inet_aton( "82.32.11.2", &mock->nextSourceAddress.sin_addr );
        mock->nextSourceAddress.sin_port = 5588;

        HeapData data;
        struct sockaddr_in sourceAddress;

        mock->recvShouldFail = true;

        TS_ASSERT_THROWS_EQUALS( rawSocket->receive( data, sourceAddress ),
                                 const Error &e,
                                 e.code(),
                                 Error::RECV_FAILED );

        TS_ASSERT_THROWS_NOTHING( delete rawSocket );
    }
};

//
// Local Variables:
// compile-command: "make -C ../.. "
// tags-file-name: "../../TAGS"
// c-basic-offset: 4
// End:
//
