#include <cxxtest/TestSuite.h>

#include "HeapData.h"
#include "MockErrno.h"
#include "RealMalloc.h"
#include "ShellCommandRunner.h"
#include "T/unistd.h"
#include "TestLog.h"

class MockForShellCommandRunner :
    public MockErrno,
    public RealMalloc,
    public T::Base_fork,
    public T::Base__exit,
    public T::Base_pipe,
    public T::Base_close,
    public T::Base_execvp,
    public T::Base_dup2,
    public T::Base_read,
    public T::Base_waitpid
{
public:
    MockForShellCommandRunner()
    {
        forkCalled = false;
        exitWasCalled = false;

        pipeWasCalled = false;
        nextPipeResult = 0;
        nextPipeReadDescriptor = 5;
        nextPipeWriteDescriptor = 43;

        lastClosedDescriptor = 0;
        nextCloseResult = 0;

        execvpWasCalled = false;

        nextDup2Result = 0;

        nextReadIndex = 0;

        readShouldFail = false;

        waitpidShouldFail = false;
        waitpidWasCalled = false;
        nextStatus = 0;
    }

    pid_t nextForkResult;
    bool forkCalled;

    pid_t fork()
    {
        TS_ASSERT( !exitWasCalled );

        forkCalled = true;
        return nextForkResult;
    }

    bool exitWasCalled;

    void _exit( int status )
    {
        TestLog::log( "Exit was called" );
        TS_ASSERT_EQUALS( status, 0 );
        exitWasCalled = true;
    }

    int nextPipeResult;
    int nextPipeReadDescriptor;
    int nextPipeWriteDescriptor;
    bool pipeWasCalled;

    int pipe( int fildes[2] )
    {
        pipeWasCalled = true;

        fildes[0] = nextPipeReadDescriptor;
        fildes[1] = nextPipeWriteDescriptor;

        return nextPipeResult;
    }

    int lastClosedDescriptor;
    int nextCloseResult;

    int close( int fd )
    {
        lastClosedDescriptor = fd;
        return nextCloseResult;
    }

    String lastExecvpFile;
    Vector<String> lastExecvpParameters;
    bool execvpWasCalled;

    int execvp( const char *file, char *const argv[] )
    {
        execvpWasCalled = true;

        lastExecvpFile = file;

        for ( unsigned i = 0; argv[i] != NULL; ++i )
        {
            lastExecvpParameters.append( argv[i] );
        }

        return -1; // If we return, it always an error...
    }

    int nextDup2Result;
    int lastDup2OldDescriptor;
    int lastDup2NewDescriptor;

    int dup2( int oldfd, int newfd )
    {
        lastDup2OldDescriptor = oldfd;
        lastDup2NewDescriptor = newfd;

        return nextDup2Result;
    }

    Vector<HeapData> nextReads;
    unsigned nextReadIndex;
    bool readShouldFail;

    ssize_t read( int fd, void *buf, size_t count )
    {
        TS_ASSERT_EQUALS( fd, nextPipeReadDescriptor );
        TS_ASSERT_EQUALS( count, (size_t)2048 );

        if ( readShouldFail )
        {
            return -1;
        }

        if ( nextReadIndex >= nextReads.size() )
        {
            return 0;
        }
        else
        {
            memcpy( buf, nextReads[nextReadIndex].data(), nextReads[nextReadIndex].size() );
            ++nextReadIndex;
            return nextReads[nextReadIndex-1].size();
        }
    }

    bool waitpidShouldFail;
    bool waitpidWasCalled;
    int nextStatus;

    pid_t waitpid( pid_t pid, int *status, int options )
    {
        waitpidWasCalled = true;

        TS_ASSERT_EQUALS( pid, nextForkResult );
        TS_ASSERT_DIFFERS( pid, 0 );

        *status = nextStatus;
        TS_ASSERT( !options );

        return waitpidShouldFail? -1 : pid;
    }
};

class ShellCommandRunnerTestSuite : public CxxTest::TestSuite
{
public:
	MockForShellCommandRunner *mock;
    int returnCode;
    String output;

    void setUp()
	{
		TS_ASSERT( mock = new MockForShellCommandRunner );
    }

	void tearDown()
	{
        TS_ASSERT_THROWS_NOTHING( delete mock );
	}

    void test_run__child()
    {
        ShellCommandRunner shellCommandRunner( "ShellCommand arg1 arg2" );

        mock->nextForkResult = 0;

        TS_ASSERT_THROWS_NOTHING( shellCommandRunner.run( returnCode, output ) );

        TS_ASSERT( mock->pipeWasCalled );

        TS_ASSERT( mock->forkCalled );

        TS_ASSERT_EQUALS( mock->lastClosedDescriptor, mock->nextPipeReadDescriptor );

        TS_ASSERT_EQUALS( mock->lastDup2OldDescriptor, mock->nextPipeWriteDescriptor );
        TS_ASSERT_EQUALS( mock->lastDup2NewDescriptor, STDOUT_FILENO );

        TS_ASSERT( !mock->waitpidWasCalled );

        TS_ASSERT( mock->execvpWasCalled );
        TS_ASSERT_EQUALS( mock->lastExecvpParameters.size(), 3U );
        TS_ASSERT_EQUALS( mock->lastExecvpParameters[0], "ShellCommand" );
        TS_ASSERT_EQUALS( mock->lastExecvpParameters[1], "arg1" );
        TS_ASSERT_EQUALS( mock->lastExecvpParameters[2], "arg2" );
    }

    void test_run__parent()
    {
        ShellCommandRunner shellCommandRunner( "ShellCommand arg1 arg2" );

        mock->nextForkResult = 117;

        String string1 = "this is the ";
        String string2 = "result of ";
        String string3 = "the execution";

        mock->nextStatus = 15;

        mock->nextReads.append( ConstSimpleData( string1.ascii(), string1.length() ) );
        mock->nextReads.append( ConstSimpleData( string2.ascii(), string2.length() ) );
        mock->nextReads.append( ConstSimpleData( string3.ascii(), string3.length() ) );

        String result;

        TS_ASSERT_THROWS_NOTHING( shellCommandRunner.run( returnCode, result ) );

        TS_ASSERT( mock->pipeWasCalled );

        TS_ASSERT( mock->forkCalled );

        TS_ASSERT_EQUALS( mock->lastClosedDescriptor, mock->nextPipeWriteDescriptor );

        TS_ASSERT( !mock->execvpWasCalled );
        TS_ASSERT( mock->waitpidWasCalled );

        TS_ASSERT_EQUALS( result, string1 + string2 + string3 );
        TS_ASSERT_EQUALS( returnCode, ( mock->nextStatus ) & 0xF0 ); // The macro used returns the 8 least significant bytes
    }

    void test_run__pipe_fails()
    {
        mock->nextPipeResult = -1;

        ShellCommandRunner shellCommandRunner( "ShellCommand arg1 arg2" );

        TS_ASSERT_THROWS_EQUALS( shellCommandRunner.run( returnCode, output ),
                                 const Error &e,
                                 e.code(),
                                 Error::PIPE_FAILED );

        TS_ASSERT( !mock->forkCalled );
        TS_ASSERT( !mock->execvpWasCalled );
    }

    void test_run__fork_fails()
    {
        ShellCommandRunner shellCommandRunner( "ShellCommand" );

        mock->nextForkResult = -1;

        TS_ASSERT_THROWS_EQUALS( shellCommandRunner.run( returnCode, output ),
                                 const Error &e,
                                 e.code(),
                                 Error::FORK_FAILED );

        TS_ASSERT( !mock->execvpWasCalled );
    }

    void test_run__child__dup_failed()
    {
        ShellCommandRunner shellCommandRunner( "ShellCommand arg1 arg2" );

        mock->nextForkResult = 0;
        mock->nextDup2Result = -1;

        TS_ASSERT_THROWS_EQUALS( shellCommandRunner.run(returnCode, output ),
                                 const Error &e,
                                 e.code(),
                                 Error::DUP_FAILED );
    }

    void test_run__parent__read_fails()
    {
        ShellCommandRunner shellCommandRunner( "ShellCommand arg1 arg2" );

        mock->nextForkResult = 117;
        mock->readShouldFail = true;

        String string1 = "this is the ";
        String string2 = "result of ";
        String string3 = "the execution";

        mock->nextReads.append( ConstSimpleData( string1.ascii(), string1.length() ) );
        mock->nextReads.append( ConstSimpleData( string2.ascii(), string2.length() ) );
        mock->nextReads.append( ConstSimpleData( string3.ascii(), string3.length() ) );

        String result;

        TS_ASSERT_THROWS_EQUALS( shellCommandRunner.run( returnCode, output ),
                                 const Error &e,
                                 e.code(),
                                 Error::READ_FAILED );
    }

    void test_run__parent__waitpid_fails()
    {
        ShellCommandRunner shellCommandRunner( "ShellCommand arg1 arg2" );

        mock->nextForkResult = 117;
        mock->waitpidShouldFail = true;

        String string1 = "this is the ";
        String string2 = "result of ";
        String string3 = "the execution";

        mock->nextReads.append( ConstSimpleData( string1.ascii(), string1.length() ) );
        mock->nextReads.append( ConstSimpleData( string2.ascii(), string2.length() ) );
        mock->nextReads.append( ConstSimpleData( string3.ascii(), string3.length() ) );

        String result;

        TS_ASSERT_THROWS_EQUALS( shellCommandRunner.run( returnCode, output ),
                                 const Error &e,
                                 e.code(),
                                 Error::WAITPID_FAILED );
    }
};

//
// Local Variables:
// compile-command: "make -C ../.. "
// tags-file-name: "../../TAGS"
// c-basic-offset: 4
// End:
//
