#ifndef __Weaver_h__
#define __Weaver_h__

#include "AutoThreadListener.h"
#include "Event.h"
#include "EventSets.h"
#include "IWeaver.h"
#include "Map.h"
#include "Oracle.h"
#include "Stringf.h"
#include "Vector.h"

class Weaver : public IWeaver
{
public:
    Weaver( ISelector &selector ) : _selector( selector )
    {
    }

    ~Weaver()
    {
        discardListeners();
    }

    void acceptRegularThread( IThreadListener *regularThreadListener )
    {
        _regularThreadListeners.append( regularThreadListener );
    }

    void acceptNonBlockingThread( IThreadListener *nonBlockingThreadListener )
    {
        _nonBlockingThreadListeners.append( nonBlockingThreadListener );
    }

    void removeThread( IThreadListener *thread )
    {
        removeThreadFromDataStructures( thread );
        destroyThread( thread );

        if ( regularThreadsSynchronized() )
        {
            selectNextEventIfPossible();
        }
    }

    void bSync( IThreadListener *thread, const Vector<Event> &requested, const Vector<Event> &blocked )
    {
        storeEvents( thread, requested, blocked );

        if ( regularThreadsSynchronized() )
        {
            selectNextEventIfPossible();
        }
    }

    void bSync( IThreadListener *thread, const Vector<Event> &requested, EventPredicate *blockPredicate )
    {
        storeEvents( thread, requested, blockPredicate );

        if ( regularThreadsSynchronized() )
        {
            selectNextEventIfPossible();
        }
    }

    void selectNextEventIfPossible()
    {
        try
        {
            selectNextEvent();
        }
        catch ( const Error &e )
        {
            if ( e.code() != Error::SELECTION_DEADLOCKED )
            {
                throw;
            }
        }
    }

private:
    ISelector &_selector;
    Vector<IThreadListener *> _regularThreadListeners;
    Vector<IThreadListener *> _nonBlockingThreadListeners;
    Map<IThreadListener *, EventSets> _threadsToEvents;
    Vector<Event> _requested;
    Vector<Event> _blocked;
    Vector<EventPredicate *> _blockPredicates;
    Vector<Event> _enabled;
    Event _selectedEvent;

    bool regularThreadsSynchronized()
    {
        for ( unsigned i = 0; i < _regularThreadListeners.size(); ++i )
        {
            if ( !_threadsToEvents.exists( _regularThreadListeners[i] ) )
            {
                return false;
            }
        }

        return true;
    }

    void storeEvents( IThreadListener *thread,
                      const Vector<Event> &requested,
                      const Vector<Event> &blocked )
    {
        _threadsToEvents[thread] = EventSets( requested, blocked );
    }

    void storeEvents( IThreadListener *thread,
                      const Vector<Event> &requested,
                      EventPredicate *blockPredicate )
    {
        _threadsToEvents[thread] = EventSets( requested, blockPredicate );
    }

    void discardListeners()
    {
        for ( unsigned i = 0; i < _regularThreadListeners.size(); ++i )
        {
            T::discardThreadListener( _regularThreadListeners[i] );
        }

        for ( unsigned i = 0; i < _nonBlockingThreadListeners.size(); ++i )
        {
            T::discardThreadListener( _nonBlockingThreadListeners[i] );
        }
    }

    void collectRequested()
    {
        Map<IThreadListener *, EventSets>::iterator it;

        for ( it = _threadsToEvents.begin(); it != _threadsToEvents.end(); ++it )
        {
            for ( unsigned i = 0; i < it->second._requested.size(); ++i )
            {
                _requested.append( it->second._requested[i] );
            }
        }
    }

    void collectBlocked()
    {
        Map<IThreadListener *, EventSets>::iterator it;

        for ( it = _threadsToEvents.begin(); it != _threadsToEvents.end(); ++it )
        {
            if ( it->second._blockedAsList )
            {
                for ( unsigned i = 0; i < it->second._blocked.size(); ++i )
                {
                    _blocked.append( it->second._blocked[i] );
                }
            }
            else
            {
                _blockPredicates.append( it->second._blockPredicate );
            }
        }
    }

    void collectEnabled()
    {
        for ( unsigned i = 0; i < _requested.size(); ++i )
        {
            if ( !_blocked.exists( _requested[i] ) && !blockedByPredicates( _requested[i] ) )
            {
                _enabled.append( _requested[i] );
            }
        }
    }

    bool blockedByPredicates( Event &event )
    {
        for ( unsigned i = 0; i < _blockPredicates.size(); ++i )
        {
            if ( _blockPredicates[i]->evaluate( event ) )
            {
                return true;
            }
        }

        return false;
    }

    void selectEvent()
    {
        collectRequested();
        collectBlocked();
        collectEnabled();

        chooseEnabled();
    }

    void chooseEnabled()
    {
        if ( _enabled.empty() )
        {
            throw Error( Error::SELECTION_DEADLOCKED );
        }

        _selectedEvent = Oracle::choose( _enabled );
    }

    void announceEvent()
    {
        for ( unsigned i = 0; i < _regularThreadListeners.size(); ++i )
        {
            _regularThreadListeners[i]->announceChosenEvent( _selectedEvent );
        }

        for ( unsigned i = 0; i < _nonBlockingThreadListeners.size(); ++i )
        {
            _nonBlockingThreadListeners[i]->announceChosenEvent( _selectedEvent );
        }

        _threadsToEvents.clear();
    }

    void clearDataStructures()
    {
        _requested.clear();
        _blocked.clear();
        _blockPredicates.clear();
        _enabled.clear();
    }

    void removeThreadFromDataStructures( IThreadListener *thread )
    {
        if ( _regularThreadListeners.exists( thread ) )
        {
            _regularThreadListeners.erase( thread );
        }
        if ( _nonBlockingThreadListeners.exists( thread ) )
        {
            _nonBlockingThreadListeners.erase( thread );
        }
        if ( _threadsToEvents.exists( thread ) )
        {
            _threadsToEvents.erase( thread );
        }
    }

    void destroyThread( IThreadListener *thread )
    {
        T::discardThreadListener( thread );
    }

    void selectNextEvent()
    {
        clearDataStructures();
        selectEvent();
        announceEvent();
    }

    void log( String line )
    {
        printf( "\t\tWeaver: %s\n", line.ascii() );
        fflush( 0 );
    }
};

#endif // __Weaver_h__

//
// Local Variables:
// compile-command: "make -C .. "
// tags-file-name: "../TAGS"
// c-basic-offset: 4
// End:
//
