#include <cxxtest/TestSuite.h>

#include "BThread.h"
#include "BlockPredicateForTests.h"
#include "Event.h"
#include "MockSelectableSocketFactory.h"
#include "MockSelectorFactory.h"
#include "MockSocketFactory.h"
#include "MockSyncPointResolverFactory.h"
#include "RealMalloc.h"

class BThreadForTests : public BThread
{
public:
    BThreadForTests() : BThread( BThread::REGULAR_THREAD )
    {
    }

    void entryPoint()
    {
    }
};

class NonBlockingBThreadForTests : public BThread
{
public:
    NonBlockingBThreadForTests() : BThread( BThread::NON_BLOCKING_THREAD )
    {
    }

    void entryPoint()
    {
    }
};

class MockForBThread :
    public RealMalloc,
    public MockSocketFactory,
    public MockSelectorFactory,
    public MockSelectableSocketFactory,
    public MockSyncPointResolverFactory
{
public:
};

class BThreadTestSuite : public CxxTest::TestSuite
{
public:
	MockForBThread *mock;
    MockSocket *socket;
    MockSelector *selector;
    MockSelectableSocket *selectableSocket;
    MockSyncPointResolver *syncPointResolver;

	void setUp()
	{
		TS_ASSERT( mock = new MockForBThread );

        socket = &( mock->mockSocket );
        selector = &( mock->mockSelector );
        selectableSocket = &( mock->mockSelectableSocket );
        syncPointResolver = &( mock->mockSyncPointResolver );
	}

	void tearDown()
	{
		TS_ASSERT_THROWS_NOTHING( delete mock );
	}

    void test_constructor_destructor()
    {
        BThreadForTests *bThread;

        TS_ASSERT( bThread = new BThreadForTests );

        TS_ASSERT( socket->wasCreatedIpPort );
        TS_ASSERT_EQUALS( socket->lastPort, EngineConfiguration::BTHREAD_MANAGEMENT_PORT );
        TS_ASSERT_EQUALS( socket->lastIp, inet_addr( EngineConfiguration::LOOPBACK_IP ) );

        TS_ASSERT( selector->wasCreated );

        TS_ASSERT( selectableSocket->wasCreated );
        TS_ASSERT_EQUALS( selectableSocket->lastSocket, socket );
        TS_ASSERT_EQUALS( selectableSocket->lastSelector, selector );

        TS_ASSERT( syncPointResolver->wasCreated );
        TS_ASSERT_EQUALS( syncPointResolver->lastSelector, selector );
        TS_ASSERT_EQUALS( syncPointResolver->lastSocket, socket );

        TS_ASSERT_THROWS_NOTHING( delete bThread );

        TS_ASSERT( socket->wasDiscarded );
        TS_ASSERT( selector->wasDiscarded );
        TS_ASSERT( syncPointResolver->wasDiscarded );
    }

    void test_constructor_destructor_nonblocking_thread()
    {
        NonBlockingBThreadForTests *bThread;

        TS_ASSERT( bThread = new NonBlockingBThreadForTests );

        TS_ASSERT( socket->wasCreatedIpPort );
        TS_ASSERT_EQUALS( socket->lastPort, EngineConfiguration::NON_BLOCKING_BTHREAD_MANAGEMENT_PORT );
        TS_ASSERT_EQUALS( socket->lastIp, inet_addr( EngineConfiguration::LOOPBACK_IP ) );

        TS_ASSERT( selector->wasCreated );

        TS_ASSERT( selectableSocket->wasCreated );
        TS_ASSERT_EQUALS( selectableSocket->lastSocket, socket );
        TS_ASSERT_EQUALS( selectableSocket->lastSelector, selector );

        TS_ASSERT( syncPointResolver->wasCreated );
        TS_ASSERT_EQUALS( syncPointResolver->lastSelector, selector );
        TS_ASSERT_EQUALS( syncPointResolver->lastSocket, socket );

        TS_ASSERT_THROWS_NOTHING( delete bThread );

        TS_ASSERT( socket->wasDiscarded );
        TS_ASSERT( selector->wasDiscarded );
        TS_ASSERT( syncPointResolver->wasDiscarded );
    }

    void test_bSync__end_in_timeout()
    {
        Event event1( 1 );

        char buffer[] = { 0x01, 0x02 };
        Event event2( 2, ConstSimpleData( buffer, sizeof(buffer) ) );

        Event event3( 3 );
        Event event4( 4 );

        BThreadForTests bThread;

        Vector<Event> requested;
        requested.append( event1 );
        requested.append( event2 );

        Vector<Event> watched;
        watched.append( event3 );

        Vector<Event> blocked;
        blocked.append( event4 );

        syncPointResolver->doneAfterNRounds = 5 + 1;
        syncPointResolver->nextLastEvent.append( IBThread::EVENT_TYPE_IF_EXPIRED );
        syncPointResolver->nextLastEventIsTimeout = true;

        String label = "somelabel";
        AtomicPropositions atomicPropositions;

        TS_ASSERT_EQUALS( bThread.bSync( requested, watched, blocked, 17, atomicPropositions, label ), IBThread::TIMEOUT );

        TS_ASSERT_EQUALS( syncPointResolver->lastRequested, requested );
        TS_ASSERT_EQUALS( syncPointResolver->lastWatched, watched );
        TS_ASSERT_EQUALS( syncPointResolver->lastBlocked, blocked );
        TS_ASSERT_EQUALS( syncPointResolver->lastTimeout, 17U );
        TS_ASSERT_EQUALS( syncPointResolver->lastLabel, label );
        TS_ASSERT( syncPointResolver->lastAtomicPropositions.empty() );

        TS_ASSERT_EQUALS( selector->numberOfSelectCalls, 5U );
    }

    void test_bSync__end_in_event()
    {
        Event event1( 1 );

        char buffer[] = { 0x01, 0x02 };
        Event event2( 2, ConstSimpleData( buffer, sizeof(buffer) ) );

        Event event3( 3 );
        Event event4( 4 );

        BThreadForTests bThread;

        Vector<Event> requested;
        requested.append( event1 );
        requested.append( event2 );

        Vector<Event> watched;
        watched.append( event3 );

        Vector<Event> blocked;
        blocked.append( event4 );

        AtomicPropositions atomicPropositions;
        atomicPropositions.add( "ap1" );
        atomicPropositions.add( "ap2" );

        syncPointResolver->doneAfterNRounds = 5 + 1;
        syncPointResolver->nextLastEvent.append( 217 );
        syncPointResolver->nextLastEventIsTimeout = false;

        String label = "somelabel";

        TS_ASSERT_EQUALS( bThread.bSync( requested, watched, blocked, 17, atomicPropositions, label ), IBThread::EVENT_WAS_TRIGGERED );

        TS_ASSERT_EQUALS( syncPointResolver->lastRequested, requested );
        TS_ASSERT_EQUALS( syncPointResolver->lastWatched, watched );
        TS_ASSERT_EQUALS( syncPointResolver->lastBlocked, blocked );
        TS_ASSERT_EQUALS( syncPointResolver->lastTimeout, 17U );
        TS_ASSERT_EQUALS( syncPointResolver->lastLabel, label );
        TS_ASSERT_EQUALS( syncPointResolver->lastAtomicPropositions.get(), atomicPropositions.get() );

        TS_ASSERT_EQUALS( selector->numberOfSelectCalls, 5U );
    }

    void test_done()
    {
        BThreadForTests *bThread;

        TS_ASSERT( bThread = new BThreadForTests );

        TS_ASSERT( !socket->closeSocketWasCalled );

        TS_ASSERT_THROWS_NOTHING( bThread->done() );
        TS_ASSERT( !socket->closeSocketWasCalled );

        TS_ASSERT_THROWS_NOTHING( delete bThread );

        TS_ASSERT( socket->wasDiscarded );
    }

    void test_bSync__end_in_event__block_predicate()
    {
        Event event1( 1 );

        char buffer[] = { 0x01, 0x02 };
        Event event2( 2, ConstSimpleData( buffer, sizeof(buffer) ) );

        Event event3( 3 );
        Event event4( 4 );

        BlockPredicateForTests blockPredicate;
        BThreadForTests bThread;

        Vector<Event> requested;
        requested.append( event1 );
        requested.append( event2 );

        Vector<Event> watched;
        watched.append( event3 );

        syncPointResolver->doneAfterNRounds = 5 + 1;
        syncPointResolver->nextLastEvent.append( 217 );
        syncPointResolver->nextLastEventIsTimeout = false;

        String label = "somelabel";

        AtomicPropositions atomicPropositions;

        TS_ASSERT_EQUALS( bThread.bSync( requested, watched, &blockPredicate, 17, atomicPropositions, label ), IBThread::EVENT_WAS_TRIGGERED );

        TS_ASSERT_EQUALS( syncPointResolver->lastRequested, requested );
        TS_ASSERT_EQUALS( syncPointResolver->lastWatched, watched );
        TS_ASSERT_EQUALS( syncPointResolver->lastBlockPredicate, &blockPredicate );
        TS_ASSERT_EQUALS( syncPointResolver->lastTimeout, 17U );
        TS_ASSERT_EQUALS( syncPointResolver->lastLabel, label );
        TS_ASSERT( syncPointResolver->lastAtomicPropositions.empty() );

        TS_ASSERT_EQUALS( selector->numberOfSelectCalls, 5U );
    }

    void test_sleep_forever()
    {
        BThreadForTests bThread;

        syncPointResolver->nextLastEvent.append( 0 );
        syncPointResolver->nextLastEventIsTimeout = false;

        TS_ASSERT_THROWS_NOTHING( bThread.sleepForever() );

        TS_ASSERT( syncPointResolver->lastRequested.empty() );
        TS_ASSERT( syncPointResolver->lastWatched.empty() );
        TS_ASSERT( syncPointResolver->lastBlocked.empty() );
        TS_ASSERT_EQUALS( syncPointResolver->lastTimeout, 0U );
        TS_ASSERT_EQUALS( syncPointResolver->lastLabel, "Asleep" );
    }
};

//
// Local Variables:
// compile-command: "make -C ../.. "
// tags-file-name: "../../TAGS"
// c-basic-offset: 4
// End:
//
