#include <cxxtest/TestSuite.h>

#include "BlockPredicateForTests.h"
#include "ConstSimpleData.h"
#include "Event.h"
#include "MockErrno.h"
#include "RealMalloc.h"
#include "SyncRequest.h"

class MockForSyncRequest :
    public RealMalloc,
    public MockErrno
{
};

class SyncRequestTestSuite : public CxxTest::TestSuite
{
public:
    MockForSyncRequest *mock;

    void setUp()
    {
        TS_ASSERT( mock = new MockForSyncRequest );
    }

    void tearDown()
    {
        TS_ASSERT_THROWS_NOTHING( delete mock );
    }

    void test_single_list_serialize()
    {
        Event event1( 0x04030201 );

        char payload[] = { 0x01, 0x02, 0x03, 0x04 };

        Event event2( 0xd4c3b2a1, ConstSimpleData( payload, sizeof(payload) ) );

        Vector<Event> requested;
        requested.append( event1 );
        requested.append( event2 );

        ListOfEvents listOfEvents( requested );

        HeapData serialized;

        TS_ASSERT_THROWS_NOTHING( listOfEvents.serialize( serialized ) );

        char expectedSerialized[] = {
            0x02, 0x00, 0x00, 0x00, // Number of events
            0x08, 0x00, 0x00, 0x00, // Length of first
            0x01, 0x02, 0x03, 0x04,
            0x00, 0x00, 0x00, 0x00,
            0x0c, 0x00, 0x00, 0x00, // Length of second
            0xa1, 0xb2, 0xc3, 0xd4, // Second
            0x04, 0x00, 0x00, 0x00,
            0x01, 0x02, 0x03, 0x04,
        };

        TS_ASSERT_EQUALS( serialized.size(), sizeof(expectedSerialized) );
        TS_ASSERT_SAME_DATA( serialized.data(), expectedSerialized, serialized.size() );
    }

    void test_sync_request_serialize__blocked_as_event_list()
    {
        char payload[] = { 0x05, 0x06, 0x07, 0x08 };
        Event event1( 0x04030201, ConstSimpleData( payload, sizeof(payload) ) );

        Event event2( 0xcdcdfefe );

        Vector<Event> requested;
        requested.append( event1 );

        Vector<Event> blocked;
        blocked.append( event2 );

        String label = "hello world";

        AtomicPropositions atomicPropositions;
        atomicPropositions.add( "ap1" );
        atomicPropositions.add( "ap2" );

        SyncRequest syncRequest( requested, blocked, atomicPropositions, label );

        HeapData serialized;

        TS_ASSERT_THROWS_NOTHING( syncRequest.serialize( serialized ) );

        char expectedSerialized[] = {
            // Label
            0x0b, 0x00, 0x00, 0x00, // Length of label
            0x68, 0x65, 0x6c, 0x6c, 0x6f, 0x20, 0x77, 0x6f, 0x72, 0x6c, 0x64,

            // Requested
            0x01, 0x00, 0x00, 0x00, // Number of events
            0x0C, 0x00, 0x00, 0x00, // Length of first
            0x01, 0x02, 0x03, 0x04, // First
            0x04, 0x00, 0x00, 0x00,
            0x05, 0x06, 0x07, 0x08,

            // Blocked
            0x00, 0x00, 0x00, 0x00,  // Type = List
            0x01, 0x00, 0x00, 0x00, // Number of events
            0x08, 0x00, 0x00, 0x00, // Length of first
            0xfe, 0xfe, 0xcd, 0xcd, // First
            0x00, 0x00, 0x00, 0x00,

            // Atomic Propositions
            0x02, 0x00, 0x00, 0x00, // Number of Properties

            0x03, 0x00, 0x00, 0x00, // Length of first
            0x61, 0x70, 0x31,       // First AP

            0x03, 0x00, 0x00, 0x00, // Length of second
            0x61, 0x70, 0x32,       // second AP
        };

        TS_ASSERT_EQUALS( serialized.size(), sizeof(expectedSerialized) );
        TS_ASSERT_SAME_DATA( serialized.data(), expectedSerialized, serialized.size() );
    }

    void test_sync_request_serialize__blocked_as_predicate()
    {
        char payload[] = { 0x05, 0x06, 0x07, 0x08 };
        Event event1( 0x04030201, ConstSimpleData( payload, sizeof(payload) ) );

        Event event2( 0xcdcdfefe );

        Vector<Event> requested;
        requested.append( event1 );

        BlockPredicateForTests blockPredicate;
        BlockPredicateForTests *predicatePointer = &blockPredicate;

        char pointerByte1 = ( ((int)predicatePointer) & 0x000000FF );
        char pointerByte2 = ( ( ((int)predicatePointer) >> 8 ) & 0x000000FF );
        char pointerByte3 = ( ( ((int)predicatePointer) >> 16 ) & 0x000000FF );
        char pointerByte4 = ( ( ((int)predicatePointer) >> 24 ) & 0x000000FF );

        String label = "hello world";

        AtomicPropositions atomicPropositions;
        atomicPropositions.add( "ap1" );
        atomicPropositions.add( "ap2" );

        SyncRequest syncRequest( requested, predicatePointer, atomicPropositions, label );

        HeapData serialized;

        TS_ASSERT_THROWS_NOTHING( syncRequest.serialize( serialized ) );

        char expectedSerialized[] = {
            // Label
            0x0b, 0x00, 0x00, 0x00, // Length of label
            0x68, 0x65, 0x6c, 0x6c, 0x6f, 0x20, 0x77, 0x6f, 0x72, 0x6c, 0x64,

            // Requested
            0x01, 0x00, 0x00, 0x00, // Number of events
            0x0C, 0x00, 0x00, 0x00, // Length of first
            0x01, 0x02, 0x03, 0x04, // First
            0x04, 0x00, 0x00, 0x00,
            0x05, 0x06, 0x07, 0x08,

            // Blocked
            0x01, 0x00, 0x00, 0x00,  // Type = Predicate
            pointerByte1, pointerByte2, pointerByte3, pointerByte4,

            // Atomic Propositions
            0x02, 0x00, 0x00, 0x00, // Number of Properties

            0x03, 0x00, 0x00, 0x00, // Length of first
            0x61, 0x70, 0x31,       // First AP

            0x03, 0x00, 0x00, 0x00, // Length of second
            0x61, 0x70, 0x32,       // second AP
        };

        TS_ASSERT_EQUALS( serialized.size(), sizeof(expectedSerialized) );
        TS_ASSERT_SAME_DATA( serialized.data(), expectedSerialized, serialized.size() );
    }

    void test_constructor_from_serialized__blocked_as_list()
    {
        char serialized[] = {
            // Label
            0x0b, 0x00, 0x00, 0x00, // Length of label
            0x68, 0x65, 0x6c, 0x6c, 0x6f, 0x20, 0x77, 0x6f, 0x72, 0x6c, 0x64,

            // Requested
            0x01, 0x00, 0x00, 0x00, // Number of events
            0x08, 0x00, 0x00, 0x00, // Length of first
            0x0a, 0x0b, 0x0c, 0x0e, // First
            0x00, 0x00, 0x00, 0x00,

            // Blocked
            0x00, 0x00, 0x00, 0x00, // Type = List
            0x01, 0x00, 0x00, 0x00, // Number of events
            0x0b, 0x00, 0x00, 0x00, // Length of first
            0x33, 0x11, 0x94, 0xa5, // First
            0x03, 0x00, 0x00, 0x00,
            0xff, 0xee, 0xdd,

            // Atomic Propositions
            0x02, 0x00, 0x00, 0x00, // Number of Properties

            0x03, 0x00, 0x00, 0x00, // Length of first
            0x61, 0x70, 0x31,       // First AP

            0x03, 0x00, 0x00, 0x00, // Length of second
            0x61, 0x70, 0x32,       // second AP
        };

        SyncRequest syncRequest( ConstSimpleData( serialized, sizeof(serialized) ) );

        TS_ASSERT_EQUALS( syncRequest.typeOfBlocked(), SyncRequest::BLOCKED_LIST );

        Vector<Event> requested = syncRequest.getRequested();
        Vector<Event> blocked = syncRequest.getBlockedEvents();
        AtomicPropositions atomicPropositions = syncRequest.getAtomicPropositions();

        TS_ASSERT_EQUALS( requested.size(), 1U );
        TS_ASSERT_EQUALS( requested[0].type(), 0x0e0c0b0aU );
        TS_ASSERT_EQUALS( requested[0].totalSize(), 8U );

        TS_ASSERT_EQUALS( blocked.size(), 1U );
        TS_ASSERT_EQUALS( blocked[0].type(), 0xa5941133U );
        TS_ASSERT_EQUALS( blocked[0].totalSize(), 11U );
        TS_ASSERT_EQUALS( blocked[0].payload().size(), 3U );

        char expectedPayload[] = { 0xff, 0xee, 0xdd };
        TS_ASSERT_SAME_DATA( blocked[0].payload().data(), expectedPayload, sizeof(expectedPayload) );

        String label = syncRequest.getLabel();
        String expectedLabel( "hello world" );

        TS_ASSERT_EQUALS( label.length(), expectedLabel.length() );
        TS_ASSERT_EQUALS( expectedLabel, label );

        TS_ASSERT_EQUALS( atomicPropositions.size(), 2U );
        TS_ASSERT_EQUALS( atomicPropositions.get()[0], "ap1" );
        TS_ASSERT_EQUALS( atomicPropositions.get()[1], "ap2" );

        TS_ASSERT_THROWS_EQUALS( syncRequest.getBlockedPredicate(),
                                 const Error &e,
                                 e.code(),
                                 Error::WRONG_TYPE_OF_BLOCKED_EVENTS );
    }

    void test_constructor_from_serialized__blocked_as_predicate()
    {
        BlockPredicateForTests blockPredicate;
        BlockPredicateForTests *predicatePointer = &blockPredicate;

        char pointerByte1 = ( ((int)predicatePointer) & 0x000000FF );
        char pointerByte2 = ( ( ((int)predicatePointer) >> 8 ) & 0x000000FF );
        char pointerByte3 = ( ( ((int)predicatePointer) >> 16 ) & 0x000000FF );
        char pointerByte4 = ( ( ((int)predicatePointer) >> 24 ) & 0x000000FF );

        char serialized[] = {
            // Label
            0x0b, 0x00, 0x00, 0x00, // Length of label
            0x68, 0x65, 0x6c, 0x6c, 0x6f, 0x20, 0x77, 0x6f, 0x72, 0x6c, 0x64,

            // Requested
            0x01, 0x00, 0x00, 0x00, // Number of events
            0x08, 0x00, 0x00, 0x00, // Length of first
            0x0a, 0x0b, 0x0c, 0x0e, // First
            0x00, 0x00, 0x00, 0x00,

            // Blocked
            0x01, 0x00, 0x00, 0x00,  // Type = Predicate
            pointerByte1, pointerByte2, pointerByte3, pointerByte4,

            // Atomic Propositions
            0x02, 0x00, 0x00, 0x00, // Number of Properties

            0x03, 0x00, 0x00, 0x00, // Length of first
            0x61, 0x70, 0x31,       // First AP

            0x03, 0x00, 0x00, 0x00, // Length of second
            0x61, 0x70, 0x32,       // second AP
        };

        SyncRequest syncRequest( ConstSimpleData( serialized, sizeof(serialized) ) );

        TS_ASSERT_EQUALS( syncRequest.typeOfBlocked(), SyncRequest::BLOCKED_PREDICATE );

        Vector<Event> requested = syncRequest.getRequested();
        AtomicPropositions atomicPropositions = syncRequest.getAtomicPropositions();

        TS_ASSERT_EQUALS( requested.size(), 1U );
        TS_ASSERT_EQUALS( requested[0].type(), 0x0e0c0b0aU );
        TS_ASSERT_EQUALS( requested[0].totalSize(), 8U );

        TS_ASSERT_EQUALS( syncRequest.getBlockedPredicate(), predicatePointer );

        String label = syncRequest.getLabel();
        String expectedLabel( "hello world" );

        TS_ASSERT_EQUALS( label.length(), expectedLabel.length() );
        TS_ASSERT_EQUALS( expectedLabel, label );

        TS_ASSERT_EQUALS( atomicPropositions.size(), 2U );
        TS_ASSERT_EQUALS( atomicPropositions.get()[0], "ap1" );
        TS_ASSERT_EQUALS( atomicPropositions.get()[1], "ap2" );

        TS_ASSERT_THROWS_EQUALS( syncRequest.getBlockedEvents(),
                                 const Error &e,
                                 e.code(),
                                 Error::WRONG_TYPE_OF_BLOCKED_EVENTS );
    }
};

//
// Local Variables:
// compile-command: "make -C ../.. "
// tags-file-name: "../../TAGS"
// c-basic-offset: 4
// End:
//
