#include <cxxtest/TestSuite.h>

#include "MockErrno.h"
#include "MockSelector.h"
#include "MockSocket.h"
#include "MockThreadListenerFactory.h"
#include "RealMalloc.h"
#include "Weaver.h"

class MockForWeaver :
    public RealMalloc,
    public MockErrno,
    public MockThreadListenerFactory
{
public:
};

class WeaverTestSuite : public CxxTest::TestSuite
{
public:
	MockForWeaver *mock;
    MockSelector *selector;

    MockThreadListener *regularListener1;
    MockThreadListener *regularListener2;
    MockThreadListener *regularListener3;

    MockThreadListener *nonBlockingListener1;
    MockThreadListener *nonBlockingListener2;
    MockThreadListener *nonBlockingListener3;

	void setUp()
	{
		TS_ASSERT( mock = new MockForWeaver );

        TS_ASSERT( selector = new MockSelector );

        TS_ASSERT( regularListener1 = new MockThreadListener );
        TS_ASSERT( regularListener2 = new MockThreadListener );
        TS_ASSERT( regularListener3 = new MockThreadListener );

        TS_ASSERT( nonBlockingListener1 = new MockThreadListener );
        TS_ASSERT( nonBlockingListener2 = new MockThreadListener );
        TS_ASSERT( nonBlockingListener3 = new MockThreadListener );

        regularListener1->wasCreated = true;
        regularListener2->wasCreated = true;
        regularListener3->wasCreated = true;

        nonBlockingListener1->wasCreated = true;
        nonBlockingListener2->wasCreated = true;
        nonBlockingListener3->wasCreated = true;
    }

	void tearDown()
	{
		TS_ASSERT_THROWS_NOTHING( delete regularListener3 );
        TS_ASSERT_THROWS_NOTHING( delete regularListener2 );
        TS_ASSERT_THROWS_NOTHING( delete regularListener1 );

		TS_ASSERT_THROWS_NOTHING( delete nonBlockingListener3 );
        TS_ASSERT_THROWS_NOTHING( delete nonBlockingListener2 );
        TS_ASSERT_THROWS_NOTHING( delete nonBlockingListener1 );

		TS_ASSERT_THROWS_NOTHING( delete selector );
        TS_ASSERT_THROWS_NOTHING( delete mock );
	}

    void test_accept_threads()
    {
        Weaver *weaver;

        TS_ASSERT( weaver = new Weaver( *selector ) );

        TS_ASSERT_THROWS_NOTHING( weaver->acceptRegularThread( regularListener1 ) );
        TS_ASSERT_THROWS_NOTHING( weaver->acceptRegularThread( regularListener2 ) );
        TS_ASSERT_THROWS_NOTHING( weaver->acceptNonBlockingThread( nonBlockingListener1 ) );
        TS_ASSERT_THROWS_NOTHING( weaver->acceptRegularThread( regularListener3 ) );
        TS_ASSERT_THROWS_NOTHING( weaver->acceptNonBlockingThread( nonBlockingListener2 ) );
        TS_ASSERT_THROWS_NOTHING( weaver->acceptNonBlockingThread( nonBlockingListener3 ) );

        TS_ASSERT_THROWS_NOTHING( delete weaver );

        TS_ASSERT( regularListener1->wasDiscarded );
        TS_ASSERT( regularListener2->wasDiscarded );
        TS_ASSERT( regularListener3->wasDiscarded );

        TS_ASSERT( nonBlockingListener1->wasDiscarded );
        TS_ASSERT( nonBlockingListener2->wasDiscarded );
        TS_ASSERT( nonBlockingListener3->wasDiscarded );
    }

    void test_bSync__blocking_works()
    {
        Weaver weaver( *selector );

        TS_ASSERT_THROWS_NOTHING( weaver.acceptRegularThread( regularListener1 ) );
        TS_ASSERT_THROWS_NOTHING( weaver.acceptRegularThread( regularListener2 ) );
        TS_ASSERT_THROWS_NOTHING( weaver.acceptRegularThread( regularListener3 ) );

        char payload[] = { 0x01, 0x02, 0x03 };

        Event event1( 1 );
        Event event2( 2 );
        Event event3( 3, ConstSimpleData( payload, sizeof(payload) ) );
        Event event4( 4 );

        Vector<Event> request1;
        Vector<Event> request2;
        Vector<Event> request3;

        Vector<Event> block1;
        Vector<Event> block2;
        Vector<Event> block3;

        request1.append( event1 );
        request1.append( event2 );

        request2.append( event2 );
        request2.append( event3 );

        block1.append( event4 );

        block3.append( event1 );
        block3.append( event2 );

        TS_ASSERT_THROWS_NOTHING( weaver.bSync( regularListener1, request1, block1 ) );
        TS_ASSERT_THROWS_NOTHING( weaver.bSync( regularListener2, request2, block2 ) );
        assertAnnounceWasntCalled();
        TS_ASSERT_THROWS_NOTHING( weaver.bSync( regularListener3, request3, block3 ) );
        assertAnnounceWasCalled( event3 );
    }

    void assertAnnounceWasCalled( Event &event )
    {
        TS_ASSERT( regularListener1->announceWasCalled );
        TS_ASSERT( regularListener2->announceWasCalled );
        TS_ASSERT( regularListener3->announceWasCalled );

        TS_ASSERT_EQUALS( regularListener1->lastAnnouncedEvent.type(), event.type() );
        TS_ASSERT_EQUALS( regularListener2->lastAnnouncedEvent.type(), event.type() );
        TS_ASSERT_EQUALS( regularListener3->lastAnnouncedEvent.type(), event.type() );

        TS_ASSERT_EQUALS( regularListener1->lastAnnouncedEvent.payload().size(), event.payload().size() );
        TS_ASSERT_EQUALS( regularListener2->lastAnnouncedEvent.payload().size(), event.payload().size() );
        TS_ASSERT_EQUALS( regularListener3->lastAnnouncedEvent.payload().size(), event.payload().size() );

        TS_ASSERT_SAME_DATA( regularListener1->lastAnnouncedEvent.payload().data(), event.payload().data(), event.payload().size() );
        TS_ASSERT_SAME_DATA( regularListener2->lastAnnouncedEvent.payload().data(), event.payload().data(), event.payload().size() );
        TS_ASSERT_SAME_DATA( regularListener3->lastAnnouncedEvent.payload().data(), event.payload().data(), event.payload().size() );

        regularListener1->announceWasCalled = false;
        regularListener2->announceWasCalled = false;
        regularListener3->announceWasCalled = false;
    }

    bool announceWasCalled( Event &event )
    {
        if ( !regularListener1->announceWasCalled || !regularListener2->announceWasCalled || !regularListener3->announceWasCalled )
        {
            return false;
        }

        if ( ( regularListener1->lastAnnouncedEvent.type() != event.type() ) ||
             ( regularListener2->lastAnnouncedEvent.type() != event.type() ) ||
             ( regularListener3->lastAnnouncedEvent.type() != event.type() ) )
        {
            return false;
        }

        if ( ( regularListener1->lastAnnouncedEvent.payload().size() != event.payload().size() ) ||
             ( regularListener2->lastAnnouncedEvent.payload().size() != event.payload().size() ) ||
             ( regularListener3->lastAnnouncedEvent.payload().size() != event.payload().size() ) )
        {
            return false;
        }

        if ( ( memcmp( regularListener1->lastAnnouncedEvent.payload().data(), event.payload().data(), event.payload().size() ) != 0 ) ||
             ( memcmp( regularListener2->lastAnnouncedEvent.payload().data(), event.payload().data(), event.payload().size() ) != 0 ) ||
             ( memcmp( regularListener3->lastAnnouncedEvent.payload().data(), event.payload().data(), event.payload().size() ) != 0 ) )
        {
            return false;
        }

        regularListener1->announceWasCalled = false;
        regularListener2->announceWasCalled = false;
        regularListener3->announceWasCalled = false;

        return true;
    }

    void assertAnnouncedToThread( MockThreadListener *thread, Event &event )
    {
        TS_ASSERT( thread->announceWasCalled );
        TS_ASSERT_EQUALS( thread->lastAnnouncedEvent.type(), event.type() );
        TS_ASSERT_EQUALS( thread->lastAnnouncedEvent.payload().size(), event.payload().size() );
        TS_ASSERT_SAME_DATA( thread->lastAnnouncedEvent.payload().data(), event.payload().data(), event.payload().size() );

        thread->announceWasCalled = false;
    }

    void assertDidntAnnounceToThread( MockThreadListener *thread )
    {
        TS_ASSERT( !thread->announceWasCalled );
    }

    void test_bSync__deadlock__then_release()
    {
        Weaver weaver( *selector );

        TS_ASSERT_THROWS_NOTHING( weaver.acceptRegularThread( regularListener1 ) );
        TS_ASSERT_THROWS_NOTHING( weaver.acceptRegularThread( regularListener2 ) );
        TS_ASSERT_THROWS_NOTHING( weaver.acceptRegularThread( regularListener3 ) );

        Event event1( 1 );
        Event event2( 2 );
        Event event3( 3 );
        Event event4( 4 );

        Vector<Event> request1;
        Vector<Event> request2;
        Vector<Event> request3;

        Vector<Event> block1;
        Vector<Event> block2;
        Vector<Event> block3;

        request1.append( event1 );
        request1.append( event2 );

        request2.append( event2 );
        request2.append( event3 );

        request3.append( event4 );

        block1.append( event4 );

        block2.append( event3 );

        block3.append( event1 );
        block3.append( event2 );

        TS_ASSERT_THROWS_NOTHING( weaver.bSync( regularListener1, request1, block1 ) );
        TS_ASSERT_THROWS_NOTHING( weaver.bSync( regularListener2, request2, block2 ) );
        assertAnnounceWasntCalled();
        TS_ASSERT_THROWS_NOTHING( weaver.bSync( regularListener3, request3, block3 ) );

        assertAnnounceWasntCalled();

        block2.clear();

        TS_ASSERT_THROWS_NOTHING( weaver.bSync( regularListener2, request2, block2 ) );

        assertAnnounceWasCalled( event3 );
    }

    void assertAnnounceWasntCalled()
    {
        TS_ASSERT( !regularListener1->announceWasCalled );
        TS_ASSERT( !regularListener2->announceWasCalled );
        TS_ASSERT( !regularListener3->announceWasCalled );
    }

    void test_bSync__announce_even_if_not_requested()
    {
        Weaver weaver( *selector );

        TS_ASSERT_THROWS_NOTHING( weaver.acceptRegularThread( regularListener1 ) );
        TS_ASSERT_THROWS_NOTHING( weaver.acceptRegularThread( regularListener2 ) );
        TS_ASSERT_THROWS_NOTHING( weaver.acceptRegularThread( regularListener3 ) );

        Event event1( 1 );
        Event event2( 2 );
        Event event3( 3 );
        Event event4( 4 );

        Vector<Event> request1;
        Vector<Event> request2;
        Vector<Event> request3;

        Vector<Event> block1;
        Vector<Event> block2;
        Vector<Event> block3;

        request1.append( event3 );
        request1.append( event2 );
        request2.append( event1 );
        request3.append( event4 );

        block2.append( event3 );
        block3.append( event1 );
        block3.append( event4 );

        TS_ASSERT_THROWS_NOTHING( weaver.bSync( regularListener1, request1, block1 ) );
        TS_ASSERT_THROWS_NOTHING( weaver.bSync( regularListener2, request2, block2 ) );
        assertAnnounceWasntCalled();
        TS_ASSERT_THROWS_NOTHING( weaver.bSync( regularListener3, request3, block3 ) );

        assertAnnounceWasCalled( event2 );
    }

    void test_bSync__consecutive_syncs()
    {
        Weaver weaver( *selector );

        TS_ASSERT_THROWS_NOTHING( weaver.acceptRegularThread( regularListener1 ) );
        TS_ASSERT_THROWS_NOTHING( weaver.acceptRegularThread( regularListener2 ) );
        TS_ASSERT_THROWS_NOTHING( weaver.acceptRegularThread( regularListener3 ) );

        Event event1( 1 );
        Event event2( 2 );
        Event event3( 3 );
        Event event4( 4 );

        Vector<Event> request1;
        Vector<Event> request2;
        Vector<Event> request3;

        Vector<Event> block1;
        Vector<Event> block2;
        Vector<Event> block3;

        block3.append( event3 );

        request1.append( event1 );
        request1.append( event2 );

        block2.append( event1 );

        TS_ASSERT_THROWS_NOTHING( weaver.bSync( regularListener1, request1, block1 ) );
        TS_ASSERT_THROWS_NOTHING( weaver.bSync( regularListener2, request2, block2 ) );
        TS_ASSERT_THROWS_NOTHING( weaver.bSync( regularListener3, request3, block3 ) );

        assertAnnounceWasCalled( event2 );

        request1.clear();

        block2.clear();
        request2.append( event2 );

        TS_ASSERT_THROWS_NOTHING( weaver.bSync( regularListener2, request2, block2 ) );
        assertAnnounceWasntCalled();

        TS_ASSERT_THROWS_NOTHING( weaver.bSync( regularListener1, request1, block1 ) );
        assertAnnounceWasntCalled();

        TS_ASSERT_THROWS_NOTHING( weaver.bSync( regularListener3, request3, block3 ) );
        assertAnnounceWasCalled( event2 );

        request2.clear();
        request2.append( event4 );

        TS_ASSERT_THROWS_NOTHING( weaver.bSync( regularListener2, request2, block2 ) );
        assertAnnounceWasntCalled();

        TS_ASSERT_THROWS_NOTHING( weaver.bSync( regularListener1, request1, block1 ) );
        assertAnnounceWasntCalled();

        TS_ASSERT_THROWS_NOTHING( weaver.bSync( regularListener3, request3, block3 ) );
        assertAnnounceWasCalled( event4 );
    }

    void test_bSync__consecutive_syncs__a_thread_is_removed()
    {
        Weaver weaver( *selector );

        TS_ASSERT_THROWS_NOTHING( weaver.acceptRegularThread( regularListener1 ) );
        TS_ASSERT_THROWS_NOTHING( weaver.acceptRegularThread( regularListener2 ) );
        TS_ASSERT_THROWS_NOTHING( weaver.acceptRegularThread( regularListener3 ) );

        Event event1( 1 );
        Event event2( 2 );
        Event event3( 3 );
        Event event4( 4 );

        Vector<Event> request1;
        Vector<Event> request2;
        Vector<Event> request3;

        Vector<Event> block1;
        Vector<Event> block2;
        Vector<Event> block3;

        block3.append( event3 );

        request1.append( event1 );
        request1.append( event2 );

        block2.append( event1 );

        TS_ASSERT_THROWS_NOTHING( weaver.bSync( regularListener1, request1, block1 ) );
        TS_ASSERT_THROWS_NOTHING( weaver.bSync( regularListener2, request2, block2 ) );
        TS_ASSERT_THROWS_NOTHING( weaver.bSync( regularListener3, request3, block3 ) );

        assertAnnounceWasCalled( event2 ) ;

        block2.clear();
        request2.append( event2 );

        TS_ASSERT_THROWS_NOTHING( weaver.bSync( regularListener2, request2, block2 ) );

        assertAnnounceWasntCalled();
        TS_ASSERT_THROWS_NOTHING( weaver.removeThread( regularListener1 ) );

        TS_ASSERT( regularListener1->wasDiscarded );
        assertAnnounceWasntCalled();

        TS_ASSERT_THROWS_NOTHING( weaver.bSync( regularListener3, request3, block3 ) );

        assertAnnouncedToThread( regularListener2, event2 );
        assertAnnouncedToThread( regularListener3, event2 );
        assertDidntAnnounceToThread( regularListener1 );
    }

    void test_bSync__blocking_works_on_events_with_payloads()
    {
        Weaver weaver( *selector );

        TS_ASSERT_THROWS_NOTHING( weaver.acceptRegularThread( regularListener1 ) );
        TS_ASSERT_THROWS_NOTHING( weaver.acceptRegularThread( regularListener2 ) );
        TS_ASSERT_THROWS_NOTHING( weaver.acceptRegularThread( regularListener3 ) );

        char payload[] = { 0x01, 0x02, 0x03 };

        Event event1( 1, ConstSimpleData( payload, sizeof(payload) ) );
        Event event2( 2 );
        Event event3( 3 );
        Event event4( 4 );

        Vector<Event> request1;
        Vector<Event> request2;
        Vector<Event> request3;

        Vector<Event> block1;
        Vector<Event> block2;
        Vector<Event> block3;

        request1.append( event1 );
        request1.append( event2 );

        request2.append( event2 );
        request2.append( event3 );

        block1.append( event4 );

        block3.append( event1 );
        block3.append( event2 );

        TS_ASSERT_THROWS_NOTHING( weaver.bSync( regularListener1, request1, block1 ) );
        TS_ASSERT_THROWS_NOTHING( weaver.bSync( regularListener2, request2, block2 ) );
        assertAnnounceWasntCalled();
        TS_ASSERT_THROWS_NOTHING( weaver.bSync( regularListener3, request3, block3 ) );
        assertAnnounceWasCalled( event3 );
    }

    void test_bSync__request_same_type_different_payloads()
    {
        Weaver weaver( *selector );

        TS_ASSERT_THROWS_NOTHING( weaver.acceptRegularThread( regularListener1 ) );
        TS_ASSERT_THROWS_NOTHING( weaver.acceptRegularThread( regularListener2 ) );
        TS_ASSERT_THROWS_NOTHING( weaver.acceptRegularThread( regularListener3 ) );

        char payload1[] = { 0x01, 0x02, 0x03 };
        char payload2[] = { 0x04, 0x05, 0x06 };
        char payload3[] = { 0x07, 0x08, 0x09 };

        Event event1( 1, ConstSimpleData( payload1, sizeof(payload1) ) );
        Event event2( 1, ConstSimpleData( payload2, sizeof(payload2) ) );
        Event event3( 1, ConstSimpleData( payload3, sizeof(payload3) ) );
        Event event4( 2 );

        Vector<Event> request1;
        Vector<Event> request2;
        Vector<Event> request3;

        Vector<Event> block1;
        Vector<Event> block2;
        Vector<Event> block3;

        request1.append( event1 );
        request1.append( event2 );

        request2.append( event3 );

        block1.append( event4 );

        TS_ASSERT_THROWS_NOTHING( weaver.bSync( regularListener1, request1, block1 ) );
        TS_ASSERT_THROWS_NOTHING( weaver.bSync( regularListener2, request2, block2 ) );
        assertAnnounceWasntCalled();
        TS_ASSERT_THROWS_NOTHING( weaver.bSync( regularListener3, request3, block3 ) );

        TS_ASSERT( announceWasCalled( event1 ) || announceWasCalled( event2 ) || announceWasCalled( event3 ) );
    }

    void test_non_blocking_threads__non_blocker_ignored()
    {
        Weaver weaver( *selector );

        TS_ASSERT_THROWS_NOTHING( weaver.acceptRegularThread( regularListener1 ) );
        TS_ASSERT_THROWS_NOTHING( weaver.acceptRegularThread( regularListener2 ) );
        TS_ASSERT_THROWS_NOTHING( weaver.acceptNonBlockingThread( nonBlockingListener3 ) );

        char payload[] = { 0x01, 0x02, 0x03 };

        Event event1( 1 );
        Event event2( 2 );
        Event event3( 3, ConstSimpleData( payload, sizeof(payload) ) );
        Event event4( 4 );

        Vector<Event> request1;
        Vector<Event> request2;
        Vector<Event> request3;

        Vector<Event> block1;
        Vector<Event> block2;
        Vector<Event> block3;

        request1.append( event1 );

        request2.append( event2 );
        request2.append( event3 );

        block1.append( event2 );
        block1.append( event3 );

        TS_ASSERT_THROWS_NOTHING( weaver.bSync( regularListener1, request1, block1 ) );

        TS_ASSERT( !regularListener1->announceWasCalled );
        TS_ASSERT( !regularListener2->announceWasCalled );
        TS_ASSERT( !nonBlockingListener3->announceWasCalled );

        assertAnnounceWasntCalled();
        TS_ASSERT_THROWS_NOTHING( weaver.bSync( regularListener2, request2, block2 ) );

        TS_ASSERT( regularListener1->announceWasCalled );
        TS_ASSERT( regularListener2->announceWasCalled );
        TS_ASSERT( nonBlockingListener3->announceWasCalled );

        TS_ASSERT_EQUALS( regularListener1->lastAnnouncedEvent.type(), 1U );
        TS_ASSERT_EQUALS( regularListener2->lastAnnouncedEvent.type(), 1U );
        TS_ASSERT_EQUALS( nonBlockingListener3->lastAnnouncedEvent.type(), 1U );
    }

    void test_non_blocking_threads__non_blocker_saves_the_day()
    {
        Weaver weaver( *selector );

        TS_ASSERT_THROWS_NOTHING( weaver.acceptRegularThread( regularListener1 ) );
        TS_ASSERT_THROWS_NOTHING( weaver.acceptRegularThread( regularListener2 ) );
        TS_ASSERT_THROWS_NOTHING( weaver.acceptNonBlockingThread( nonBlockingListener3 ) );

        char payload[] = { 0x01, 0x02, 0x03 };

        Event event1( 1 );
        Event event2( 2 );
        Event event3( 3, ConstSimpleData( payload, sizeof(payload) ) );
        Event event4( 4 );

        Vector<Event> request1;
        Vector<Event> request2;
        Vector<Event> request3;

        Vector<Event> block1;
        Vector<Event> block2;
        Vector<Event> block3;

        request1.append( event1 );
        request2.append( event2 );

        block1.append( event2 );
        block2.append( event1 );

        request3.append( event4 );

        TS_ASSERT_THROWS_NOTHING( weaver.bSync( regularListener1, request1, block1 ) );

        TS_ASSERT( !regularListener1->announceWasCalled );
        TS_ASSERT( !regularListener2->announceWasCalled );
        TS_ASSERT( !nonBlockingListener3->announceWasCalled );

        assertAnnounceWasntCalled();
        TS_ASSERT_THROWS_NOTHING( weaver.bSync( regularListener2, request2, block2 ) );

        TS_ASSERT( !regularListener1->announceWasCalled );
        TS_ASSERT( !regularListener2->announceWasCalled );
        TS_ASSERT( !nonBlockingListener3->announceWasCalled );

        TS_ASSERT_THROWS_NOTHING( weaver.bSync( regularListener3, request3, block3 ) );

        TS_ASSERT( regularListener1->announceWasCalled );
        TS_ASSERT( regularListener2->announceWasCalled );
        TS_ASSERT( nonBlockingListener3->announceWasCalled );

        TS_ASSERT_EQUALS( regularListener1->lastAnnouncedEvent.type(), 4U );
        TS_ASSERT_EQUALS( regularListener2->lastAnnouncedEvent.type(), 4U );
        TS_ASSERT_EQUALS( nonBlockingListener3->lastAnnouncedEvent.type(), 4U );
    }

    void test_bSync__blocking_works__using_predicates()
    {
        Weaver weaver( *selector );

        TS_ASSERT_THROWS_NOTHING( weaver.acceptRegularThread( regularListener1 ) );
        TS_ASSERT_THROWS_NOTHING( weaver.acceptRegularThread( regularListener2 ) );
        TS_ASSERT_THROWS_NOTHING( weaver.acceptRegularThread( regularListener3 ) );

        char payloadOne[] = { 0x01, 0x02, 0x03 };
        char payloadTwo[] = { 0x01, 0x02, 0x03, 0x04 };

        Event event1( 1 );
        Event event2( 2 );
        Event event3a( 3, ConstSimpleData( payloadOne, sizeof(payloadOne) ) );
        Event event3b( 3, ConstSimpleData( payloadTwo, sizeof(payloadTwo) ) );
        Event event4( 4 );

        Vector<Event> request1;
        Vector<Event> request2;
        Vector<Event> request3;

        Vector<Event> block1;
        Vector<Event> block2;
        Vector<Event> block3;

        request1.append( event1 );
        request1.append( event2 );

        request2.append( event2 );
        request2.append( event3a );
        request2.append( event3b );

        class MyBlock : public EventPredicate
        {
        public:
            bool evaluate( const Event &event )
            {
                if ( event.type() != 3 )
                {
                    return true;
                }
                if ( event.payloadSize() < 4 )
                {
                    return true;
                }

                return false;
            }
        };

        MyBlock myBlock;

        TS_ASSERT_THROWS_NOTHING( weaver.bSync( regularListener1, request1, block1 ) );
        TS_ASSERT_THROWS_NOTHING( weaver.bSync( regularListener2, request2, block2 ) );
        assertAnnounceWasntCalled();
        TS_ASSERT_THROWS_NOTHING( weaver.bSync( regularListener3, request3, &myBlock ) );
        assertAnnounceWasCalled( event3b );
    }

};

//
// Local Variables:
// compile-command: "make -C ../.. "
// tags-file-name: "../../TAGS"
// c-basic-offset: 4
// End:
//
