#ifndef __CustomizedOracleChooser_h__
#define __CustomizedOracleChooser_h__

#include "IOracle.h"
#include "IpEvent.h"
#include "TcpSegment.h"

class CustomizedOracleChooser : public IOracle::OracleChooser
{
public:
    Event choose( Vector<Event> &enabledEvents )
    {
        Vector<Event> lowPriority;
        Vector<Event> mediumPriority;
        Vector<Event> highPriority;

        occassionallyClearOldConnections();

        for ( unsigned i = 0; i < enabledEvents.size(); ++i )
        {
            if ( enabledEvents[i].type() == IEvent::IDLE_EVENT )
            {
                lowPriority.append( enabledEvents[i] );
            }
            else if ( enabledEvents[i].type() == IEvent::TCP_FIN_REQUEST )
            {
                mediumPriority.append( enabledEvents[i] );
            }
            else
            {
                highPriority.append( enabledEvents[i] );
            }
        }

        if ( !highPriority.empty() )
        {
            return favorStarvingConnections( highPriority );
        }
        else if ( !mediumPriority.empty() )
        {
            return randomEvent( mediumPriority );
        }
        else
        {
            return randomEvent( lowPriority );
        }
    }

private:
    Vector<ConnectionIdentifier> _existingConnections;

    void occassionallyClearOldConnections()
    {
        if ( _existingConnections.size() >= 50 )
        {
            _existingConnections.clear();
        }
    }

    Event randomEvent( Vector<Event> &candidates )
    {
        return candidates[rand() % candidates.size()];
    }

    Event favorStarvingConnections( Vector<Event> &candidates )
    {
        if ( !candidates.exists( IEvent::TCP_OUTGOING_SEGMENT ) )
        {
            return randomEvent( candidates );
        }
        else
        {
            Vector<ConnectionIdentifier> connectionIdentifiers = gatherConnectionIdentifiers( candidates );
            ConnectionIdentifier winner = chooseStarvingConnection( connectionIdentifiers );
            updateDatabase( winner );
            log( Stringf ( "Have TCP_OUTGOING_SEGMENT. Winner is port: %u", winner._port ) );
            return eventByIdentifier( winner, candidates );
        }
    }

    void updateDatabase( ConnectionIdentifier winner )
    {
        if ( _existingConnections.exists( winner ) )
        {
            _existingConnections.eraseByValue( winner );
        }
        _existingConnections.insertHead( winner );
    }

    Event eventByIdentifier( ConnectionIdentifier identifier, Vector<Event> &candidates )
    {
        for ( unsigned i = 0; i < candidates.size(); ++i )
        {
            if ( ( candidates[i].type() == IEvent::TCP_OUTGOING_SEGMENT ) &&
                 ( identifier == extractConnectionIdentifier( candidates[i] ) ) )
            {
                return candidates[i];
            }
        }
        throw Error( Error::CONNECTION_IDENTIFIER_NOT_FOUND );
    }

    UniqueVector<ConnectionIdentifier> gatherConnectionIdentifiers( Vector<Event> &candidates )
    {
        UniqueVector<ConnectionIdentifier> connectionIdentifiers;

        for ( unsigned i = 0; i < candidates.size(); ++i )
        {
            if ( candidates[i].type() == IEvent::TCP_OUTGOING_SEGMENT )
            {
                connectionIdentifiers.append( extractConnectionIdentifier( candidates[i] ) );
            }
        }

        return connectionIdentifiers;
    }

    ConnectionIdentifier extractConnectionIdentifier( Event event )
    {
        IpEvent ipEvent;
        ipEvent.constructFromEvent( event );

        TcpSegment tcpSegment;
        tcpSegment.unserialize( ipEvent.getPacketPayload() );

        return ConnectionIdentifier( ipEvent.getIp(), tcpSegment.header().getDestinationPort() );
    }

    ConnectionIdentifier chooseStarvingConnection( Vector<ConnectionIdentifier> &connections )
    {
        if ( connections.empty() )
        {
            throw Error( Error::NO_CONNECTION_IDENTIFIERS );
        }

        ConnectionIdentifier currentWinner = connections.first();

        if ( !_existingConnections.exists( currentWinner ) )
        {
            return currentWinner;
        }

        for ( unsigned i = 1; i < connections.size(); ++i )
        {
            if ( !_existingConnections.exists( connections[i] ) )
            {
                return connections[i];
            }
            else if ( _existingConnections.find( currentWinner ) < _existingConnections.find( connections[i] ) )
            {
                currentWinner = connections[i];
            }
        }

        return currentWinner;
    }

    void log( String message )
    {
        return;
        printf( "CustomizedOracleChooser:\t%s\n", message.ascii() );
        fflush( 0 );
    }
};

#endif // __CustomizedOracleChooser_h__

//
// Local Variables:
// compile-command: "make -C .. "
// tags-file-name: "../TAGS"
// c-basic-offset: 4
// End:
//
