#ifndef __HttpGetHandler_h__
#define __HttpGetHandler_h__

#include "BThread.h"
#include "File.h"
#include "IEvent.h"
#include "ServerConfiguration.h"

class HttpGetHandler : public BThread
{
public:
    HttpGetHandler() : BThread( BThread::REGULAR_THREAD )
    {
    }

    void entryPoint()
    {
        try
        {
            while ( true )
            {
                waitForHttpGet();
                storeRequest();
                extractFileName();

                if ( fileDoesntExist() )
                {
                    log( "File doesnt exist\n" );
                    sendErrorMessage();
                }
                else if ( directoryNotEndingWithSlash() )
                {
                    log( "Directory not ending with slash\n" );
                    sendFoundMessage();
                }
                else if ( cgiRequest() )
                {
                    log ( "Cgi file requested\n" );
                    handleCgiRequest();
                }

                else
                {
                    considerDefaultFileName();
                    sendResponse();
                }
            }
        }
        catch ( Error &e )
        {
            e.setUserMessage( "Thread: HttpGetHandler" );
            throw e;
        }
    }

private:
    enum {
        MAX_CHUNK_SIZE = 4096
    };

    String _fileName;
    String _fileNameFullPath;
    TcpEvent _getRequest;

    void waitForHttpGet()
    {
        clearEventVectors();

        _watched.append( IEvent::HTTP_GET_REQUEST );
        bSync( _requested, _watched, _blocked, IBThread::NO_TIMEOUT );
    }

    void storeRequest()
    {
        _getRequest.constructFromEvent( lastEvent() );
    }

    void extractFileName()
    {
        String requestString( _getRequest.getPacketPayload().asChar(), _getRequest.getPacketPayload().size() );
        _fileName = requestString.tokenize( " " )[1];
        fixSpaceEncoding();

        if ( _fileName.contains( "?" ) )
        {
            _fileName = _fileName.tokenize( "?" )[0];
        }

        _fileNameFullPath = String( ServerConfiguration::LOCAL_DIRECTORY ) + _fileName;

        log( Stringf( "\t    GET: %s\n", _fileNameFullPath.ascii() ) );
    }

    void fixSpaceEncoding()
    {
        static const char WRONG_ENCODING[] = "%20";
        static const char CORRECT_ENCODING[] = " ";

        while ( _fileName.contains( WRONG_ENCODING ) )
        {
            _fileName.replace( WRONG_ENCODING, CORRECT_ENCODING );
        }
    }

    bool directoryNotEndingWithSlash()
    {
        return File::directory( _fileNameFullPath ) && ( _fileNameFullPath[_fileNameFullPath.length() - 1] != '/' );
    }

    void considerDefaultFileName()
    {
        if ( File::directory( _fileNameFullPath ) )
        {
            appendDefaultSuffix();
        }
    }

    void appendDefaultSuffix()
    {
        static char DEFAULT_FILE_NAME[] = "index.html";
        _fileNameFullPath += DEFAULT_FILE_NAME;
    }

    bool fileDoesntExist()
    {
        return !File::exists( _fileNameFullPath );
    }

    String contentType()
    {
        String type;

        if ( _fileNameFullPath.contains( ".html" ) )
        {
            type = "text/html; charset=iso-8859-1";
        }
        else if ( _fileNameFullPath.contains( ".css" ) )
        {
            type = "text/css; charset=iso-8859-1";
        }
        else if ( _fileNameFullPath.contains( ".js" ) )
        {
            type = "application/x-javascript";
        }
        else
        {
            type = "text/plain; charset=iso-8859-1";
        }

        return Stringf( "Content-Type: %s\r\n", type.ascii() );
    }

    void sendResponse()
    {
        sendHeader();
        sendContent();
    }

    void sendHeader()
    {
        Stringf headerString( "%s%s%s%s%u%s",
                              "HTTP/1.1 200 Ok\r\n",
                              "Server: BP Server\r\n",
                              contentType().ascii(),
                              "Content-Length: ",
                              File::getSize( _fileNameFullPath ),
                              "\r\nConnection: close\r\n\r\n" );

        HeapData header = ConstSimpleData( headerString.ascii(), headerString.length() );

        clearEventVectors();
        _requested.append( TcpEvent( IEvent::DATA_FROM_HTTP_TO_TCP, _getRequest.getIp(), _getRequest.getPort(), header ) );
        _blocked.append( IEvent::HTTP_GET_REQUEST );
        bSync( _requested, _watched, _blocked, IBThread::NO_TIMEOUT );
    }

    void sendContent()
    {
        File inputFile( _fileNameFullPath );
        inputFile.open( IFile::MODE_READ );

        unsigned indexInFile = 0;

        while ( indexInFile < File::getSize( _fileNameFullPath ) )
        {
            sendChunk( inputFile );
            indexInFile += MAX_CHUNK_SIZE;
        }
    }

    void sendChunk( IFile &file )
    {
        HeapData content;
        file.read( content, MAX_CHUNK_SIZE );

        clearEventVectors();
        _requested.append( TcpEvent( IEvent::DATA_FROM_HTTP_TO_TCP, _getRequest.getIp(), _getRequest.getPort(), content ) );
        _blocked.append( IEvent::HTTP_GET_REQUEST );
        bSync( _requested, _watched, _blocked, IBThread::NO_TIMEOUT );
    }

    void sendErrorMessage()
    {
        ConstSimpleData empty( 0, 0 );

        clearEventVectors();
        _requested.append( TcpEvent( IEvent::HTTP_GENERATE_ERROR, _getRequest.getIp(), _getRequest.getPort(), empty ) );
        _blocked.append( IEvent::HTTP_GET_REQUEST );
        bSync( _requested, _watched, _blocked, IBThread::NO_TIMEOUT );
    }

    void sendFoundMessage()
    {
        clearEventVectors();
        _requested.append( TcpEvent( IEvent::HTTP_GENERATE_FOUND, _getRequest.getIp(), _getRequest.getPort(), ConstSimpleData( _fileName.ascii(), _fileName.length() ) ) );
        _blocked.append( IEvent::HTTP_GET_REQUEST );
        bSync( _requested, _watched, _blocked, IBThread::NO_TIMEOUT );
    }

    bool cgiRequest() const
    {
        return _fileNameFullPath.contains( ServerConfiguration::CGI_PATH ) && _fileNameFullPath.endsWith( ".cgi" );
    }

    void handleCgiRequest()
    {
        clearEventVectors();
        _requested.append( TcpEvent( IEvent::HTTP_GET_CGI_REQUEST, _getRequest.getIp(), _getRequest.getPort(), _getRequest.getPacketPayload() ) );
        _blocked.append( IEvent::HTTP_GET_REQUEST );
        bSync( _requested, _watched, _blocked, IBThread::NO_TIMEOUT );
    }

    void log( String message )
    {
        return;
        printf( "HttpGetHandler:\t%s\n", message.ascii() );
    }
};

#endif // __HttpGetHandler_h__

//
// Local Variables:
// compile-command: "make -C .. "
// tags-file-name: "../TAGS"
// c-basic-offset: 4
// End:
//
