#ifndef __PeriodicSender_h__
#define __PeriodicSender_h__

#include "ConnectionIdentifier.h"
#include "IpEvent.h"
#include "ServerConfiguration.h"
#include "SpawnableBThread.h"
#include "Time.h"

class PeriodicSender : public SpawnableBThread
{
public:
    PeriodicSender( ConnectionIdentifier connectionIdentifier, const ConstSimpleData &packet, unsigned lastByte ) :
        SpawnableBThread( BThread::REGULAR_THREAD ), _connectionIdentifier( connectionIdentifier ), _packet( packet ), _lastByte( lastByte ), _ackReceived( false )
    {
    }

    void entryPoint()
    {
        fixDeadline();

        while ( !ackReceived() )
        {
            waitForAck();

            if ( !ackReceived() && deadlinePassed() )
            {
                log( "Resending packet" );
                resendPacket();
                fixDeadline();
            }
        }

        done();
    }

private:
    ConnectionIdentifier _connectionIdentifier;
    HeapData _packet;
    unsigned _lastByte;
    bool _ackReceived;
    time_t _deadline;

    void fixDeadline()
    {
        _deadline = Time::sample() + ServerConfiguration::SECONDS_FOR_RETRANSMISSON;
    }

    bool deadlinePassed() const
    {
        return Time::sample() > _deadline;
    }

    bool ackReceived() const
    {
        return _ackReceived;
    }

    void resendPacket()
    {
        clearEventVectors();

        IpEvent request( IEvent::TCP_OUTGOING_SEGMENT, _connectionIdentifier.ipAsString(), _packet );
        _requested.append( request );
        _blocked.append( IEvent::TCP_ACK_REQUEST );
        bSync( _requested, _watched, _blocked, IBThread::NO_TIMEOUT );
    }

    void waitForAck()
    {
        clearEventVectors();
        _watched.append( IEvent::TCP_ACK_REQUEST );
        if ( bSync( _requested, _watched, _blocked, ServerConfiguration::SECONDS_FOR_RETRANSMISSON ) == IBThread::EVENT_WAS_TRIGGERED )
        {
            processAck();
        }
    }

    void processAck()
    {
        TcpEvent tcpEvent;
        tcpEvent.constructFromEvent( lastEvent() );

        TcpSegment tcpSegment;
        tcpSegment.unserialize( tcpEvent.getPacketPayload() );

        ConnectionIdentifier connectionIdentifier( tcpEvent.getIp(), tcpEvent.getPort() );

        if ( connectionIdentifier != _connectionIdentifier )
        {
            return;
        }

        unsigned ackedByte = tcpSegment.header().getAckNumber();

        if ( _lastByte <= ackedByte )
        {
            _ackReceived = true;
        }
    }

    void log( String message )
    {
        return;
        printf( "Single Packet Retransmitter: %s\n", message.ascii() );
    }
};

#endif // __PeriodicSender_h__

//
// Local Variables:
// compile-command: "make -C .. "
// tags-file-name: "../TAGS"
// c-basic-offset: 4
// End:
//
