#ifndef __SequenceMaster_h__
#define __SequenceMaster_h__

#include "BThread.h"
#include "ConnectionIdentifier.h"
#include "IEvent.h"
#include "Map.h"
#include "SequenceNumberPayload.h"

class SequenceMaster : public BThread
{
public:
    SequenceMaster() : BThread( BThread::REGULAR_THREAD )
    {
    }

    void entryPoint()
    {
        try
        {
            while ( true )
            {
                waitForSequenceNumberRequest();
                storeRequest();
                fetchSequenceNumber();
                sendSequenceNumber();
            }
        }
        catch ( Error &e )
        {
            e.setUserMessage( "Thread: SequenceMaster" );
            throw e;
        }
    }

private:
    unsigned _nextSequenceNumber;
    unsigned _sizeOfNextPacket;
    Map<ConnectionIdentifier, unsigned> _connectionToSequenceNumber;
    TcpEvent _request;

    Vector<Event> allRequestTypes()
    {
        Vector<Event> types;

        types.append( IEvent::ACK_SENDER_REQUEST_SEQUENCE_NUMBER );
        types.append( IEvent::SYN_HANDLER_REQUEST_SEQUENCE_NUMBER );
        types.append( IEvent::FIN_HANDLER_REQUEST_SEQUENCE_NUMBER );
        types.append( IEvent::PUSH_SENDER_REQUEST_SEQUENCE_NUMBER );
        types.append( IEvent::FIN_SENDER_REQUEST_SEQUENCE_NUMBER );
        types.append( IEvent::KEEPALIVE_SENDER_REQUEST_SEQUENCE_NUMBER );

        return types;
    }

    Vector<Event> allResponseTypes()
    {
        Vector<Event> types;

        types.append( IEvent::SEQUENCE_NUMBER_FOR_ACK_SENDER );
        types.append( IEvent::SEQUENCE_NUMBER_FOR_SYN_HANDLER );
        types.append( IEvent::SEQUENCE_NUMBER_FOR_FIN_HANDLER );
        types.append( IEvent::SEQUENCE_NUMBER_FOR_PUSH_SENDER );
        types.append( IEvent::SEQUENCE_NUMBER_FOR_FIN_SENDER );
        types.append( IEvent::SEQUENCE_NUMBER_FOR_KEEPALIVE_SENDER );

        return types;
    }

    void waitForSequenceNumberRequest()
    {
        clearEventVectors();

        _watched = allRequestTypes();
        _blocked = allResponseTypes();

        bSync( _requested, _watched, _blocked, IBThread::NO_TIMEOUT );
    }

    void storeRequest()
    {
        _request.constructFromEvent( lastEvent() );
    }

    void fetchSequenceNumber()
    {
        ConnectionIdentifier connectionIdentifier( _request.getIp(), _request.getPort() );

        if ( !_connectionToSequenceNumber.exists( connectionIdentifier ) )
        {
            _connectionToSequenceNumber[connectionIdentifier] = rand();
        }

        _nextSequenceNumber = _connectionToSequenceNumber[connectionIdentifier];

        _connectionToSequenceNumber[connectionIdentifier] += SequenceNumberPayload( _request.getPacketPayload() ).getBytesToBeSent();
    }

    void sendSequenceNumber()
    {
        enum {
            DONT_CARE = 0,
        };

        clearEventVectors();
        blockAllRelevantEvents();

        SequenceNumberPayload payload( _nextSequenceNumber, DONT_CARE );

        switch ( lastEvent().type() )
        {
        case IEvent::ACK_SENDER_REQUEST_SEQUENCE_NUMBER:
            _requested.append( TcpEvent( IEvent::SEQUENCE_NUMBER_FOR_ACK_SENDER, _request.getIp(), _request.getPort(), ConstSimpleData( &payload, sizeof(payload) ) ) );
            _blocked.eraseByValue( IEvent::SEQUENCE_NUMBER_FOR_ACK_SENDER );
            break;

        case IEvent::SYN_HANDLER_REQUEST_SEQUENCE_NUMBER:
            _requested.append( TcpEvent( IEvent::SEQUENCE_NUMBER_FOR_SYN_HANDLER, _request.getIp(), _request.getPort(), ConstSimpleData( &payload, sizeof(payload) ) ) );
            _blocked.eraseByValue( IEvent::SEQUENCE_NUMBER_FOR_SYN_HANDLER );
            break;

        case IEvent::FIN_HANDLER_REQUEST_SEQUENCE_NUMBER:
            _requested.append( TcpEvent( IEvent::SEQUENCE_NUMBER_FOR_FIN_HANDLER, _request.getIp(), _request.getPort(), ConstSimpleData( &payload, sizeof(payload) ) ) );
            _blocked.eraseByValue( IEvent::SEQUENCE_NUMBER_FOR_FIN_HANDLER );
            break;

        case IEvent::PUSH_SENDER_REQUEST_SEQUENCE_NUMBER:
            _requested.append( TcpEvent( IEvent::SEQUENCE_NUMBER_FOR_PUSH_SENDER, _request.getIp(), _request.getPort(), ConstSimpleData( &payload, sizeof(payload) ) ) );
            _blocked.eraseByValue( IEvent::SEQUENCE_NUMBER_FOR_PUSH_SENDER );
            break;

        case IEvent::FIN_SENDER_REQUEST_SEQUENCE_NUMBER:
            _requested.append( TcpEvent( IEvent::SEQUENCE_NUMBER_FOR_FIN_SENDER, _request.getIp(), _request.getPort(), ConstSimpleData( &payload, sizeof(payload) ) ) );
            _blocked.eraseByValue( IEvent::SEQUENCE_NUMBER_FOR_FIN_SENDER );
            break;

        case IEvent::KEEPALIVE_SENDER_REQUEST_SEQUENCE_NUMBER:
            _requested.append( TcpEvent( IEvent::SEQUENCE_NUMBER_FOR_KEEPALIVE_SENDER, _request.getIp(), _request.getPort(), ConstSimpleData( &payload, sizeof(payload) ) ) );
            _blocked.eraseByValue( IEvent::SEQUENCE_NUMBER_FOR_KEEPALIVE_SENDER );
            break;

        default:
            break;
        }

        bSync( _requested, _watched, _blocked, IBThread::NO_TIMEOUT );
    }

    void blockAllRelevantEvents()
    {
        _blocked = allRequestTypes();
        Vector<Event> responses = allResponseTypes();
        _blocked += responses;
    }
};

#endif // __SequenceMaster_h__

//
// Local Variables:
// compile-command: "make -C .. "
// tags-file-name: "../TAGS"
// c-basic-offset: 4
// End:
//
