#ifndef __TcpFinHandler_h__
#define __TcpFinHandler_h__

#include "BThread.h"
#include "ConnectionIdentifier.h"
#include "IEvent.h"
#include "ServerConfiguration.h"

class TcpFinHandler : public BThread
{
public:
    TcpFinHandler() : BThread( BThread::REGULAR_THREAD )
    {
    }

    void entryPoint()
    {
        try
        {
            while ( true )
            {
                waitForTcpFinRequests();

                if ( sendFinRequest() )
                {
                    markFinAsSent();
                }
                else
                {
                    storeFinRequest();
                    getFreshSequenceNumber();
                    updateAckNumber();
                    replyWithFinAck();
                }
            }
        }
        catch ( Error &e )
        {
            e.setUserMessage( "Thread: TcpFinHandler" );
            throw e;
        }
    }

private:
    unsigned _sequenceNumber;
    TcpEvent _finRequest;
    TcpSegment _tcpSegment;
    Map<ConnectionIdentifier, unsigned char> _previouslySentFins;

    void waitForTcpFinRequests()
    {
        clearEventVectors();

        _watched.append( IEvent::TCP_FIN_REQUEST );
        _watched.append( IEvent::TCP_SEND_FIN );

        _blocked.append( IEvent::FIN_HANDLER_REQUEST_SEQUENCE_NUMBER );
        _blocked.append( IEvent::SEQUENCE_NUMBER_FOR_FIN_HANDLER );

        bSync( _requested, _watched, _blocked, IBThread::NO_TIMEOUT );
    }

    bool sendFinRequest()
    {
        return lastEvent().type() == IEvent::TCP_SEND_FIN;
    }

    bool alreadySentFin()
    {
        return _previouslySentFins.exists( ConnectionIdentifier( _finRequest.getIp(), _finRequest.getPort() ) );
    }

    void markFinAsSent()
    {
        TcpEvent finRequest;
        finRequest.constructFromEvent( lastEvent() );
        _previouslySentFins[ConnectionIdentifier( finRequest.getIp(), finRequest.getPort() )] = 0;
    }

    void storeFinRequest()
    {
        _finRequest.constructFromEvent( lastEvent() );
        _tcpSegment.unserialize( _finRequest.getPacketPayload() );
    }

    void updateAckNumber()
    {
        AckNumberPayload payload( _tcpSegment.header().getSequenceNumber() + 1 );
        clearEventVectors();
        _requested.append( TcpEvent( IEvent::SET_ACK_NUMBER, _finRequest.getIp(), _finRequest.getPort(), ConstSimpleData( &payload, sizeof(payload) ) ) );
        _blocked = allRelevantEventsExcept( IEvent::SET_ACK_NUMBER );
        bSync( _requested, _watched, _blocked, IBThread::NO_TIMEOUT );
    }

    void replyWithFinAck()
    {
        TcpSegment finAck;
        prepareFinAckSegment( finAck );
        sendAck( finAck );
    }

    void prepareFinAckSegment( TcpSegment &finAck )
    {
        finAck.header().setSourcePort( _tcpSegment.header().getDestinationPort() );
        finAck.header().setDestinationPort( _tcpSegment.header().getSourcePort() );

        finAck.header().setAckNumber( _tcpSegment.header().getSequenceNumber() + _tcpSegment._payload.size() + 1 );
        //        finAck.header().setAckNumber( _tcpSegment.header().getSequenceNumber() + 1 );
        finAck.header().setSequenceNumber( _sequenceNumber );

        finAck.header().setAck();

        if ( !alreadySentFin() )
        {
            finAck.header().setFin();
            markFinAsSent();
        }

        finAck.header().setWindowSize( ServerConfiguration::TCP_WINDOW_SIZE );
    }

    void sendAck( TcpSegment &finAck )
    {
        HeapData serialized;
        finAck.serialize( serialized, ServerConfiguration::SERVER_IP, _finRequest.getIp() );

        clearEventVectors();
        IpEvent finAckEvent( IEvent::TCP_OUTGOING_SEGMENT, _finRequest.getIp(), serialized );
        _requested.append( finAckEvent );

        _blocked = allRelevantEventsExcept( IEvent::TCP_OUTGOING_SEGMENT );

        bSync( _requested, _watched, _blocked, IBThread::NO_TIMEOUT );
    }

    void getFreshSequenceNumber()
    {
        requestSequenceNumber();
        readSequenceNumber();
    }

    void requestSequenceNumber()
    {
        enum {
            DONT_CARE = 0,
            FIN_REQUEST_COUNTS_AS_ONE_BYTE = 1,
        };

        clearEventVectors();

        SequenceNumberPayload sequenceNumber( DONT_CARE, FIN_REQUEST_COUNTS_AS_ONE_BYTE );

        _requested.append( TcpEvent( IEvent::FIN_HANDLER_REQUEST_SEQUENCE_NUMBER,
                                     _finRequest.getIp(),
                                     _finRequest.getPort(),
                                     ConstSimpleData( &sequenceNumber, sizeof(sequenceNumber) ) ) );
        _blocked = allRelevantEventsExcept( IEvent::FIN_HANDLER_REQUEST_SEQUENCE_NUMBER );
        bSync( _requested, _watched, _blocked, IBThread::NO_TIMEOUT );
    }

    void readSequenceNumber()
    {
        clearEventVectors();

        _watched.append( IEvent::SEQUENCE_NUMBER_FOR_FIN_HANDLER );
        _blocked = allRelevantEventsExcept( IEvent::SEQUENCE_NUMBER_FOR_FIN_HANDLER );
        bSync( _requested, _watched, _blocked, IBThread::NO_TIMEOUT );

        TcpEvent sequenceNumber;
        sequenceNumber.constructFromEvent( lastEvent() );

        _sequenceNumber = SequenceNumberPayload( sequenceNumber.getPacketPayload() ).getSequenceNumber();
    }

    Vector<Event> allRelevantEventsExcept( Event event )
    {
        Vector<Event> result;

        if ( event != IEvent::TCP_FIN_REQUEST ) result.append( IEvent::TCP_FIN_REQUEST );
        if ( event != IEvent::TCP_SEND_FIN ) result.append( IEvent::TCP_FIN_REQUEST );
        if ( event != IEvent::FIN_HANDLER_REQUEST_SEQUENCE_NUMBER ) result.append( IEvent::FIN_HANDLER_REQUEST_SEQUENCE_NUMBER );
        if ( event != IEvent::SEQUENCE_NUMBER_FOR_FIN_HANDLER ) result.append( IEvent::SEQUENCE_NUMBER_FOR_FIN_HANDLER );

        return result;
    }
};

#endif // __TcpFinHandler_h__

//
// Local Variables:
// compile-command: "make -C .. "
// tags-file-name: "../TAGS"
// c-basic-offset: 4
// End:
//
