#ifndef __TcpFinSender_h__
#define __TcpFinSender_h__

#include "BThread.h"
#include "IEvent.h"
#include "ServerConfiguration.h"

class TcpFinSender : public BThread
{
public:
    TcpFinSender() : BThread( BThread::REGULAR_THREAD )
    {
    }

    void entryPoint()
    {
        try
        {
            while ( true )
            {
                waitForSendTcpFin();
                storeSendFinRequest();
                getFreshSequenceNumber();
                sendFin();
            }
        }
        catch ( Error &e )
        {
            e.setUserMessage( "Thread: TcpFinSender" );
            throw e;
        }
    }

private:
    unsigned _sequenceNumber;
    TcpEvent _sendFinRequest;

    void waitForSendTcpFin()
    {
        clearEventVectors();

        _watched.append( Event( IEvent::TCP_SEND_FIN ) );
        _blocked = allRelevantEventsExcept( IEvent::TCP_SEND_FIN );
        bSync( _requested, _watched, _blocked, IBThread::NO_TIMEOUT );
    }

    void storeSendFinRequest()
    {
        _sendFinRequest.constructFromEvent( lastEvent() );
    }

    void sendFin()
    {
        TcpSegment fin;
        prepareFinSegment( fin );
        sendFinSegment( fin );
    }

    void prepareFinSegment( TcpSegment &fin )
    {
        fin.header().setSourcePort( ServerConfiguration::LISTENING_PORT );
        fin.header().setDestinationPort( _sendFinRequest.getPort() );

        fin.header().setSequenceNumber( _sequenceNumber );

        fin.header().setFin();

        fin.header().setWindowSize( ServerConfiguration::TCP_WINDOW_SIZE );
    }

    void sendFinSegment( TcpSegment &fin )
    {
        HeapData serialized;
        fin.serialize( serialized, ServerConfiguration::SERVER_IP, _sendFinRequest.getIp() );

        clearEventVectors();
        IpEvent finEvent( IEvent::TCP_OUTGOING_SEGMENT, _sendFinRequest.getIp(), serialized );
        _requested.append( finEvent );

        _blocked = allRelevantEventsExcept( IEvent::TCP_OUTGOING_SEGMENT );

        bSync( _requested, _watched, _blocked, IBThread::NO_TIMEOUT );
    }

    void getFreshSequenceNumber()
    {
        requestSequenceNumber();
        readSequenceNumber();
    }

    void requestSequenceNumber()
    {
        enum {
            DONT_CARE = 0,
            FIN_REQUEST_COUNTS_AS_ONE_BYTE = 1,
        };

        clearEventVectors();

        SequenceNumberPayload sequenceNumber( DONT_CARE, FIN_REQUEST_COUNTS_AS_ONE_BYTE );

        _requested.append( TcpEvent( IEvent::FIN_SENDER_REQUEST_SEQUENCE_NUMBER,
                                     _sendFinRequest.getIp(),
                                     _sendFinRequest.getPort(),
                                     ConstSimpleData( &sequenceNumber, sizeof(sequenceNumber) ) ) );
        _blocked = allRelevantEventsExcept( IEvent::FIN_SENDER_REQUEST_SEQUENCE_NUMBER );
        bSync( _requested, _watched, _blocked, IBThread::NO_TIMEOUT );
    }

    void readSequenceNumber()
    {
        clearEventVectors();

        _watched.append( IEvent::SEQUENCE_NUMBER_FOR_FIN_SENDER );
        _blocked = allRelevantEventsExcept( IEvent::SEQUENCE_NUMBER_FOR_FIN_SENDER );
        bSync( _requested, _watched, _blocked, IBThread::NO_TIMEOUT );

        TcpEvent sequenceNumber;
        sequenceNumber.constructFromEvent( lastEvent() );

        _sequenceNumber = SequenceNumberPayload( sequenceNumber.getPacketPayload() ).getSequenceNumber();
    }

    Vector<Event> allRelevantEventsExcept( Event event )
    {
        Vector<Event> result;

        if ( event != IEvent::TCP_SEND_FIN ) result.append( IEvent::TCP_SEND_FIN );
        if ( event != IEvent::FIN_SENDER_REQUEST_SEQUENCE_NUMBER ) result.append( IEvent::FIN_SENDER_REQUEST_SEQUENCE_NUMBER );
        if ( event != IEvent::SEQUENCE_NUMBER_FOR_FIN_SENDER ) result.append( IEvent::SEQUENCE_NUMBER_FOR_FIN_SENDER );

        return result;
    }
};

#endif // __TcpFinSender_h__

//
// Local Variables:
// compile-command: "make -C .. "
// tags-file-name: "../TAGS"
// c-basic-offset: 4
// End:
//
