#ifndef __TcpSegment_h__
#define __TcpSegment_h__

#include "HeapData.h"
#include "ServerConfiguration.h"
#include "String.h"
#include <arpa/inet.h>

struct PseudoHeader
{
public:
    enum {
        TCP_PROTOCOL = 6,
    };

    PseudoHeader() : _protocol( TCP_PROTOCOL )
    {
        memset( &_zeroes, 0, sizeof(_zeroes) );
    }

    unsigned long _sourceIp;
    unsigned long _destinationIp;
    unsigned char _zeroes;
    unsigned char _protocol;
    unsigned short _tcpLength;
} __attribute((packed));

struct TcpHeader
{
public:
    enum {
        DATA_OFFSET_MASK = 0xf0,
        FIN_BIT_MASK = 0x01,
        SYN_BIT_MASK = 0x02,
        ACK_BIT_MASK = 0x10,
        PUSH_BIT_MASK = 0x08,
        BYTES_IN_LONG = 4,
    };

    TcpHeader()
    {
        memset( &_mandatoryPart, 0, sizeof(_mandatoryPart) );
    }

    struct MandatoryPart
    {
        unsigned short _sourcePort;
        unsigned short _destinationPort;
        unsigned _sequenceNumber;
        unsigned _ackNumber;
        unsigned char _dataOffsetNsFlag;
        unsigned char _flags;
        unsigned short _windowSize;
        unsigned short _checkSum;
        unsigned short _urgentPointer;
    } __attribute((packed));

    MandatoryPart _mandatoryPart;
    HeapData _options;

    bool getSyn() const
    {
        return _mandatoryPart._flags & SYN_BIT_MASK;
    }

    void setSyn()
    {
        _mandatoryPart._flags |= SYN_BIT_MASK;
    }

    bool getFin() const
    {
        return _mandatoryPart._flags & FIN_BIT_MASK;
    }

    void setFin()
    {
        _mandatoryPart._flags |= FIN_BIT_MASK;
    }

    void setAck()
    {
        _mandatoryPart._flags |= ACK_BIT_MASK;
    }

    bool getAck() const
    {
        return _mandatoryPart._flags & ACK_BIT_MASK;
    }

    bool getPush() const
    {
        return _mandatoryPart._flags & PUSH_BIT_MASK;
    }

    void setPush()
    {
        _mandatoryPart._flags |= PUSH_BIT_MASK;
    }

    unsigned getTcpHeaderLength() const
    {
        unsigned dataOffset = ( _mandatoryPart._dataOffsetNsFlag >> 4 ) & 0x0f;
        return dataOffset * BYTES_IN_LONG;
    }

    void setTcpHeaderLength()
    {
        unsigned length = sizeof(_mandatoryPart) + _options.size();
        unsigned char lengthAsChar = ( length / BYTES_IN_LONG );
        lengthAsChar = ( lengthAsChar << 4 ) & DATA_OFFSET_MASK;

        _mandatoryPart._dataOffsetNsFlag |= lengthAsChar;
    }

    void setWindowSize( unsigned short windowSize )
    {
        _mandatoryPart._windowSize = windowSize;
    }

    void unserialize( const ConstSimpleData &serialized )
    {
        memcpy( &_mandatoryPart, serialized.data(), sizeof(MandatoryPart) );

        _options = ConstSimpleData( serialized.asChar() + sizeof(_mandatoryPart), getTcpHeaderLength() - sizeof(_mandatoryPart) );

        _mandatoryPart._sourcePort      = ntohs( _mandatoryPart._sourcePort );
        _mandatoryPart._destinationPort = ntohs( _mandatoryPart._destinationPort );
        _mandatoryPart._sequenceNumber  = ntohl( _mandatoryPart._sequenceNumber );
        _mandatoryPart._ackNumber       = ntohl( _mandatoryPart._ackNumber );
        _mandatoryPart._windowSize      = ntohs( _mandatoryPart._windowSize );
        _mandatoryPart._urgentPointer   = ntohs( _mandatoryPart._urgentPointer );
    }

    void serialize( HeapData &serialized )
    {
        setTcpHeaderLength();

        MandatoryPart copyOfMandatoryPart = _mandatoryPart;

        copyOfMandatoryPart._sourcePort      = htons( copyOfMandatoryPart._sourcePort );
        copyOfMandatoryPart._destinationPort = htons( copyOfMandatoryPart._destinationPort );
        copyOfMandatoryPart._sequenceNumber  = htonl( copyOfMandatoryPart._sequenceNumber );
        copyOfMandatoryPart._ackNumber       = htonl( copyOfMandatoryPart._ackNumber );
        copyOfMandatoryPart._windowSize      = htons( copyOfMandatoryPart._windowSize );
        copyOfMandatoryPart._urgentPointer   = htons( copyOfMandatoryPart._urgentPointer );

        serialized = ConstSimpleData( &copyOfMandatoryPart, sizeof(copyOfMandatoryPart) );
        serialized += _options;
    }

    unsigned short getSourcePort() const
    {
        return _mandatoryPart._sourcePort;
    }

    unsigned short getDestinationPort() const
    {
        return _mandatoryPart._destinationPort;
    }

    void setSourcePort( unsigned short port )
    {
        _mandatoryPart._sourcePort = port;
    }

    void setDestinationPort( unsigned short port )
    {
        _mandatoryPart._destinationPort = port;
    }

    unsigned getSequenceNumber() const
    {
        return _mandatoryPart._sequenceNumber;
    }

    unsigned getAckNumber() const
    {
        return _mandatoryPart._ackNumber;
    }

    void setSequenceNumber( unsigned sequenceNumber )
    {
        _mandatoryPart._sequenceNumber = sequenceNumber;
    }

    void setAckNumber( unsigned ackNumber )
    {
        _mandatoryPart._ackNumber = ackNumber;
    }
};

class TcpSegment
{
public:
    TcpHeader _header;
    HeapData _payload;

    void serialize( HeapData &serialized, String sourceIp, String destinationIp )
    {
        _header.setTcpHeaderLength();
        setChecksum( sourceIp, destinationIp );

        _header.serialize( serialized );
        serialized += _payload;
    }

    void unserialize( const ConstSimpleData &serialized )
    {
        _header.unserialize( serialized );
        _payload = ConstSimpleData( serialized.asChar() + _header.getTcpHeaderLength(), serialized.size() - _header.getTcpHeaderLength() );
    }

    TcpHeader &header()
    {
        return _header;
    }

    static unsigned short computeChecksum( unsigned short *buffer, int length )
    {
        long sum = 0;

        for ( ; length > 1; length -= 2, ++buffer )
        {
            sum += *buffer;
        }

        if ( length > 0 )
        {
            sum += *(unsigned char *)buffer;
        }

        while ( sum >> 16 )
        {
            sum = ( sum & 0xffff ) + ( sum >> 16 );
        }

        sum = ~sum;

        return (unsigned short)sum;
    }

    void setChecksum( String sourceIp, String destinationIp )
    {
        PseudoHeader pseudoHeader;
        preparePseudoHeader( pseudoHeader, sourceIp, destinationIp );
        addChecksum( pseudoHeader );
    }

    void preparePseudoHeader( PseudoHeader &pseudoHeader, String sourceIp, String destinationIp )
    {
        unsigned char sourceIpAsBytes[4];
        unsigned short sourceIpAsShorts[4];
        unsigned char destinationIpAsBytes[4];
        unsigned short destinationIpAsShorts[4];

        sscanf( sourceIp.ascii(), "%hu.%hu.%hu.%hu", sourceIpAsShorts, sourceIpAsShorts + 1, sourceIpAsShorts + 2, sourceIpAsShorts + 3 );
        sscanf( destinationIp.ascii(), "%hu.%hu.%hu.%hu", destinationIpAsShorts, destinationIpAsShorts + 1, destinationIpAsShorts + 2, destinationIpAsShorts + 3 );

        for ( unsigned i = 0; i < 4; ++i )
        {
            sourceIpAsBytes[i] = (unsigned char)sourceIpAsShorts[i];
            destinationIpAsBytes[i] = (unsigned char)destinationIpAsShorts[i];
        }

        memcpy( &pseudoHeader._sourceIp, sourceIpAsBytes, sizeof(pseudoHeader._sourceIp) );
        memcpy( &pseudoHeader._destinationIp, destinationIpAsBytes, sizeof(pseudoHeader._destinationIp) );

        pseudoHeader._tcpLength = htons( _header.getTcpHeaderLength() + _payload.size() );
    }

    void addChecksum( PseudoHeader &pseudoHeader )
    {
        zeroCheckSum();

        HeapData buffer = ConstSimpleData( &pseudoHeader, sizeof(pseudoHeader) );

        HeapData temp;
        _header.serialize( temp );
        buffer += temp;

        buffer += _payload;

        _header._mandatoryPart._checkSum = computeChecksum( (unsigned short *)buffer.data(), buffer.size() );
    }

    void zeroCheckSum()
    {
        _header._mandatoryPart._checkSum = 0;
    }
};

#endif // __TcpSegment_h__

//
// Local Variables:
// compile-command: "make -C .. "
// tags-file-name: "../TAGS"
// c-basic-offset: 4
// End:
//
