#ifndef __TcpSynHandler_h__
#define __TcpSynHandler_h__

#include "BThread.h"
#include "IEvent.h"
#include "ServerConfiguration.h"
#include "TargetPortPayload.h"

class TcpSynHandler : public BThread
{
public:
    TcpSynHandler() : BThread( BThread::REGULAR_THREAD )
    {
    }

    void entryPoint()
    {
        try
        {
            while ( true )
            {
                waitForTcpSynRequest();
                storeSynRequest();
                getFreshSequenceNumber();
                updateAckNumber();
                replyWithSynAck();
            }
        }
        catch ( Error &e )
        {
            e.setUserMessage( "Thread: TcpSynHandler" );
            throw e;
        }
    }

private:
    unsigned _sequenceNumber;
    TcpEvent _synRequest;
    TcpSegment _tcpSegment;

    void storeSynRequest()
    {
        _synRequest.constructFromEvent( lastEvent() );
        _tcpSegment.unserialize( _synRequest.getPacketPayload() );
    }

    void waitForTcpSynRequest()
    {
        clearEventVectors();

        _watched.append( IEvent::TCP_SYN_REQUEST );
        _blocked = allRelevantEventsExcept( IEvent::TCP_SYN_REQUEST );
        bSync( _requested, _watched, _blocked, IBThread::NO_TIMEOUT );

        log( "Got syn request" );
    }

    void updateAckNumber()
    {
        AckNumberPayload payload( _tcpSegment.header().getSequenceNumber() + 1 );

        clearEventVectors();
        _requested.append( TcpEvent( IEvent::SET_ACK_NUMBER, _synRequest.getIp(), _synRequest.getPort(), ConstSimpleData( &payload, sizeof(payload) ) ) );
        _blocked = allRelevantEventsExcept( IEvent::SET_ACK_NUMBER );
        bSync( _requested, _watched, _blocked, IBThread::NO_TIMEOUT );
    }

    void replyWithSynAck()
    {
        TcpSegment synAck;
        prepareAckSegment( synAck );
        sendSynAck( synAck );
    }

    void prepareAckSegment( TcpSegment &synAck )
    {
        synAck.header().setSourcePort( _tcpSegment.header().getDestinationPort() );
        synAck.header().setDestinationPort( _tcpSegment.header().getSourcePort() );

        log( Stringf( "Replying to port: %u", _tcpSegment.header().getSourcePort() ) );

        synAck.header().setAckNumber( _tcpSegment.header().getSequenceNumber() + 1 );
        synAck.header().setSequenceNumber( _sequenceNumber );

        synAck.header().setAck();
        synAck.header().setSyn();

        synAck.header().setWindowSize( ServerConfiguration::TCP_WINDOW_SIZE );
    }

    void sendSynAck( TcpSegment &synAck )
    {
        HeapData serialized;
        synAck.serialize( serialized, ServerConfiguration::SERVER_IP, _synRequest.getIp() );

        clearEventVectors();
        IpEvent synAckEvent( IEvent::TCP_OUTGOING_SEGMENT, _synRequest.getIp(), serialized );
        _requested.append( synAckEvent );

        _blocked = allRelevantEventsExcept( IEvent::TCP_OUTGOING_SEGMENT );

        bSync( _requested, _watched, _blocked, IBThread::NO_TIMEOUT );
    }

    void getFreshSequenceNumber()
    {
        requestSequenceNumber();
        readSequenceNumber();
    }

    void requestSequenceNumber()
    {
        enum {
            DONT_CARE = 0,
            SYN_REQUEST_COUNTS_AS_ONE_BYTE = 1,
        };

        clearEventVectors();

        SequenceNumberPayload sequenceNumber( DONT_CARE, SYN_REQUEST_COUNTS_AS_ONE_BYTE );

        _requested.append( TcpEvent( IEvent::SYN_HANDLER_REQUEST_SEQUENCE_NUMBER,
                                     _synRequest.getIp(),
                                     _synRequest.getPort(),
                                     ConstSimpleData( &sequenceNumber, sizeof(sequenceNumber) ) ) );
        _blocked = allRelevantEventsExcept( IEvent::SYN_HANDLER_REQUEST_SEQUENCE_NUMBER );
        bSync( _requested, _watched, _blocked, IBThread::NO_TIMEOUT );
    }

    void readSequenceNumber()
    {
        clearEventVectors();

        _watched.append( IEvent::SEQUENCE_NUMBER_FOR_SYN_HANDLER );
        _blocked = allRelevantEventsExcept( IEvent::SEQUENCE_NUMBER_FOR_SYN_HANDLER );
        bSync( _requested, _watched, _blocked, IBThread::NO_TIMEOUT );

        TcpEvent sequenceNumber;
        sequenceNumber.constructFromEvent( lastEvent() );

        _sequenceNumber = SequenceNumberPayload( sequenceNumber.getPacketPayload() ).getSequenceNumber();
    }

    Vector<Event> allRelevantEventsExcept( Event event )
    {
        Vector<Event> result;

        if ( event != IEvent::TCP_SYN_REQUEST ) result.append( IEvent::TCP_SYN_REQUEST );
        if ( event != IEvent::SYN_HANDLER_REQUEST_SEQUENCE_NUMBER ) result.append( IEvent::SYN_HANDLER_REQUEST_SEQUENCE_NUMBER );
        if ( event != IEvent::SEQUENCE_NUMBER_FOR_SYN_HANDLER ) result.append( IEvent::SEQUENCE_NUMBER_FOR_SYN_HANDLER );

        return result;
    }

    void log( String message )
    {
        return;
        printf( "TcpSynHandler:\t%s\n", message.ascii() );
    }
};

#endif // __TcpSynHandler_h__

//
// Local Variables:
// compile-command: "make -C .. "
// tags-file-name: "../TAGS"
// c-basic-offset: 4
// End:
//
